/* ----------------------------------------------------------------- */
/* PMprintf -- Server for PRINTF queued strings                      */
/* Copyright (c) IBM Corporation, 1991, 1992                         */
/* ----------------------------------------------------------------- */
/*                                     Mike Cowlishaw, Jan/Feb 1991  */
/*                                     32-bit          Jan/Feb 1992  */
/*                                     Andreas Buchinger  2009-2024  */

/*------ Make constant string definition more efficient & protect them! -----*/
#pragma  strings(readonly)

/*--- Include header files --------------------------------------------------*/
#define INCL_WIN
#define INCL_WINSTDFILE
#define INCL_DOS
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#define INCL_TYPES
#define INCL_DOSSEMAPHORES
#define INCL_LONGLONG  // 20090705 AB
#define INCL_LOADEXCEPTQ

#include <os2.h>
#include "exceptq.h"
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <time.h>
#include <stdlib.h>
#include <ctype.h>
#include "pfexe.h"
#include "FeCommon.h"
#include "question.h"

#include <float.h>

/*--- General Global Variables ----------------------------------------------*/
#define MAX_CLASS_NAME_LENGTH		52 		// 20240101 found out by a lot of frustating trial and error and idebug sessions

#define DEFAULT_NUMBER_OF_LISTBOX_ITEMS         5000
#define DEFAULT_NUMBER_OF_LISTBOX_ITEMS_STRING "5000"
static HAB            hab;
static HMQ            hmq;
static CMDLINEOPTIONS CmdLineOptions      = DEFAULT_CMDLINEOPTIONS;
static char    NameOfLogFile[CCHMAXPATH]  = "";
static PFNWP   ListBoxFunction            = NULL;
static BOOL    BeepOnMessageArrival       = FALSE;
static ULONG   NumberMessagesInNamedQueue = 0;
static char    FullQueueName[CCHMAXPATH]  = "";
#define QUEUE_NAME_PREFIX  "\\QUEUES\\"
static BOOL    LoggingToFile              = FALSE;
static BOOL    SavePresParms              = TRUE;
static BOOL    RestorePresParms           = TRUE;
static BOOL    LogTimeWithMsgToFile       = TRUE;
static BOOL    AddTimeToDisplayedLines    = TRUE;
static ULONG   MaxLinesKeptInWindow       = DEFAULT_NUMBER_OF_LISTBOX_ITEMS;
static HEV     hSemPaused                 = 0;
static HEV     hSemGotLine                = 0;
static LONG    LinesInListbox             = 0;
static LONG    LinesVisibleInListbox;
static HFILE   LogfileHandle;
static BOOL    QueuePaused                = FALSE;
static BOOL    CaseSensitiveSearches      = FALSE;
static char    SearchString[100]          = "";
static SHORT   LastMatchLocation          = LIT_NONE;
static char    LogTimeFormat[50+1]        = DEFAULT_TIME_FMT;

/*--- Window & Menu Handles -------------------------------------------------*/
static HWND   hFrame;
static HWND   hClient;
static HWND   hContextMenu = NULLHANDLE;
static HWND   hApplicationMenu;
static HWND   hListBox;
#define DEFAULT_LISTBOX_FONT  "8.Helv"
static HWND   hHelpInstance;


/*--- User messages ---------------------------------------------------------*/
#define UM_CREATED                  WM_USER+0 /* window ready */
#define UM_SIZE                     WM_USER+1 /* new size please */
#define UM_QUEUE_CREATED            WM_USER+2
#define UM_ERROR_QUEUE_DUPLICATED   WM_USER+3
#define UM_ERROR_CANT_CREATE_QUEUE  WM_USER+4
#define UM_ERROR_CANT_READ_QUEUE    WM_USER+5  //DosRc is passed
#define UM_GOTLINE                    WM_USER+6  /* [addr;len,bell] line received */
#define UM_ERROR_WRITING_TO_LOG_FILE  WM_USER+8
#define UM_FILETRUNC                  WM_USER+9  /* truncation while writing file */






/*===========================================================================*/
static void _Optlink StatusBarString(  ULONG   BeepFrequency,
#define BEEP_ERROR 4000L
																			 char  * Text
																		)
/*===========================================================================*/						
	{
	char   StatusBar[200];

	if ( *Text == '\0' )
		StatusBar[0] = '\0';
	else
		sprintf(StatusBar, "%.150s", Text);
	WinSendMsg(hApplicationMenu, MM_SETITEMTEXT, MPFROMSHORT(IDM_STATUS_BAR), MPFROMP(StatusBar));

	/*--- Beep if this was requested -----------------------------------------*/
	if ( BeepFrequency )
		DosBeep(BeepFrequency, 80L);
	}


/*===========================================================================*/
static void _Optlink StatusBarClear(void)
/*===========================================================================*/
	{
	StatusBarString(FALSE, "");
	}

/*===========================================================================*/
static void _Optlink StatusBarMessage(ULONG BeepFrequency, char  * Fmt, ...)
/*===========================================================================*/
	{
	char    StatusBar[CCHMAXPATH + 100 + 1];
	char  * String;

	if ( *Fmt == '\0' )
		String = Fmt;
	else
		{
		va_list   Args;

		/*-- Format the message ----------------------------------------------*/
		va_start(Args, Fmt);
		vsprintf(StatusBar, Fmt, Args);
		va_end(Args);
		String = StatusBar;
		}
	StatusBarString(BeepFrequency, String);
	}




/*==========================================================================*/
static ULONG _Optlink MsgBoxVa(
															ULONG   PopupAttribute,
#define ATTR_INFO  MB_INFORMATION
#define ATTR_WINFO MB_NOICON
#define ATTR_ERROR MB_ERROR
															char  * Title,
															char  * Format,
															va_list Args
															)
/*==========================================================================*/
	{
	char    * MessageBuffer;
#define LARGEST_POPUPMSG_IN_CHARS 2000
	ULONG     Rc;
	char    * Message;

	/*--- Allocate memory for the message ------------------------------------*/
	if ( (MessageBuffer=malloc(LARGEST_POPUPMSG_IN_CHARS+1)) == NULL )
		{
		StatusBarMessage(BEEP_ERROR, "MsgBoxVa() could not allocate memory...");
		return(MB_CANCEL);
		}

	/*--- Remove the status message ------------------------------------------*/
	StatusBarClear();

	/*-- Format the message --------------------------------------------------*/
	vsprintf(MessageBuffer, Format, Args);

	/*--- If the title is NULL then the 1st line of message is it! -----------*/
	if ( Title != NULL )
		Message = MessageBuffer;				//We were passed the title
	else
		{
		/*--- The 1st line is the title (find end of line and terminate) -----*/
		Title = MessageBuffer;
		if ( (Message=strchr(MessageBuffer, '\n')) == NULL )
			Message    = "";							 //Only one line!
		else
			*Message++ = '\0';						 //null terminate title & point to message
		}

	/*-- Display the message -------------------------------------------------*/
	Rc = WinMessageBox(  HWND_DESKTOP,
											 hFrame,
											 Message,
											 Title,
											 0,
											 PopupAttribute | MB_MOVEABLE
										);

	/*--- Free the allocated memory ------------------------------------------*/
	free(MessageBuffer);
	return(Rc);
	}



/*==========================================================================*/
static void _Optlink MsgBox(
													 ULONG   PopupAttribute,
													 char  * Title,
													 char  * Format,
													 ...
													 )
/*==========================================================================*/
	{
	va_list   Args;

	/*-- Display the message -------------------------------------------------*/
	va_start(Args, Format);
	MsgBoxVa(PopupAttribute, Title, Format, Args);
	va_end(Args);
	}



/*===========================================================================*/
static char * _Optlink strIchr(const char *Haystack, int NeedleChar)
/*===========================================================================*/
	{
	const char  * HaystackPtr;

	/*--- Look for the passed character -------------------------------------*/
	NeedleChar = toupper(NeedleChar);
	for ( HaystackPtr=Haystack; *HaystackPtr != '\0'; HaystackPtr++ )
		{
		/*--- Try this position! --------------------------------------------*/
		if ( toupper(*HaystackPtr) == NeedleChar )
			return((char *)HaystackPtr);					 //Found the character
		}

	/*--- Character was not found -------------------------------------------*/
	return(NULL);
	}



/*===========================================================================*/
static char * _Optlink strIstr(const char *Haystack, const char *Needle)
						/*                                                                           */
						/* This is a case insensitive 'C' strstr() function.                         */
/*===========================================================================*/
	{
	int     NeedlesLength = strlen(Needle);
	const char  * HaystackPtr;
	char    Needles1stChar;

	/*--- Return Haystack if Needle contain contains nothing ----------------*/
	if ( *Needle == '\0' )
		return((char *)Haystack);

	/*--- Search Haystack for 1st byte of Needle ----------------------------*/
	Needles1stChar = toupper(*Needle);
	for ( HaystackPtr=Haystack; (HaystackPtr=strIchr(HaystackPtr, Needles1stChar)) != NULL; HaystackPtr++ )
		{
		/*--- We have found a match on the first byte, see if rest OK -------*/
		if ( strnicmp(HaystackPtr, Needle, NeedlesLength) == 0 )
			return((char *)HaystackPtr);								//String matches
		}

	/*--- Sub string not found in string ------------------------------------*/
	return(NULL);
	}




/*===========================================================================*/
static char  * _Optlink FmtShortenFileName(char  *ShorterResult, char  *FileName, int RequiredLength)
						/*                                                                           */
						/*    Pull the guts out of a filename that is longer than that specified     */
						/*    to produce something like :-  "C:\..\TURBO\SOURCE\NAME.PAS"            */
						/*                                                                           */
						/*    NOTE: This routine expects to be passed a complete filename, that is   */
						/*    one that includes a drive & path.                                      */
/*===========================================================================*/
	{
	int       Length = strlen(FileName);
	char    * Ptr;

	/*---- Either leave as is (may be short name) or shorten the name ------*/
	if ( Length > RequiredLength )
		{
		if ( RequiredLength < 20 )
			{
			*ShorterResult = '\0';							 /* Make result zero length */
			return("Too Short, should be at least 20 bytes");
			}
		*ShorterResult++ = FileName[0];						/* [0] Get Drive Letter             */
		*ShorterResult++ = FileName[1];						/* [1] Get the ":"                  */
		*ShorterResult++ = FileName[2];						/* [2] Get the "\"                  */
		*ShorterResult++ = '\xFA';								/* [3] Add the "." (1st)            */
		*ShorterResult++ = '\xFA';								/* [4] Add the "." (2nd)            */
		Ptr = &FileName[Length-RequiredLength+5];	/* Point to the first possible byte */
		while ( (*Ptr != '\\') && (*Ptr != 0) )
			Ptr++;																/* Find the next "\"                */
		strcpy(ShorterResult, Ptr);								/* Copy rest of the filename        */
		}
	else
		strcpy(ShorterResult, FileName);					 /* Its short enough already         */

	return(ShorterResult);												 /* Return a pointer to the result   */
	}



/*===========================================================================*/
static void _Optlink StartWpsObject(char  * WpsObjectId, char  * Description)
						/*                                                                           */
						/* Although a better method exists under warp, I wish this program to work   */
						/* as well as possible in older 32 bit releases (2.0+).                      */
/*===========================================================================*/
	{
	ULONG   ObjectsHandle;

	/*--- Get the objects handle given the ID --------------------------------*/
	if ( (ObjectsHandle=WinQueryObject(WpsObjectId)) == NULLHANDLE )
		{
		/*--- Object does not exist ------------------------------------------*/
		MsgBox(ATTR_ERROR, NULL,
					 "%s DOES NOT EXIST\n"
					 "The required palette object does not appear to exist."
					 "\n\n"
					 "If you are not using OS/2 WARP or later then that "
					 "is probably the reason (the font palette should work).  "
					 "You may wish to start the palette objects yourself, they normally "
					 "exist in the \"System Setup\" folder."
					 ,Description
					);
		return;
		}

	/*--- Now start the bloody thing -----------------------------------------*/
	if ( WinSetObjectData(ObjectsHandle, "OPEN=DEFAULT") == FALSE )
		{
		MsgBox(ATTR_ERROR, NULL,
					 "CAN'T START %s\n"
					 "The palette exists but we could not start it..."
					 ,Description
					);
		}
	}





/*===========================================================================*/
static void _Optlink SavePositionAndPresentationParameters(BOOL Force)
/*===========================================================================*/
	{
	/*--- Save details -------------------------------------------------------*/
	if ( SavePresParms || Force )
		WinStoreWindowPos(APP_PROFILE_NAME, KEY_WINDOW_PARMS_AND_POSN, hFrame);
	}



/*===========================================================================*/
static void _Optlink RestorePositionAndPresentationParameters(HWND Handle)
/*===========================================================================*/
	{
	SWP Swp;

	if ( RestorePresParms )
		{
		/*--- Restore the window ---------------------------------------------*/
		WinRestoreWindowPos(APP_PROFILE_NAME, KEY_WINDOW_PARMS_AND_POSN, Handle);

		/*--- If minimized we need to restore the window! --------------------*/
		if ( WinQueryWindowPos(Handle, &Swp) && (BitsAreSet(Swp.fl, SWP_MINIMIZE)) )
			{
			/*--- Restore the window -----------------------------------------*/
			WinSetWindowPos(Handle, NULLHANDLE, 0,0, 0,0, SWP_RESTORE);
			}
		}
	}



/*===========================================================================*/
static ULONG _Optlink GetProfileUlong(char  * KeyName, ULONG DefaultValue)
/*===========================================================================*/
	{
	ULONG FromIni;
	ULONG DataLength = sizeof(FromIni);

	/*--- Get the value from the INI file ------------------------------------*/
	if ( PrfQueryProfileData(HINI_USERPROFILE, APP_PROFILE_NAME, KeyName, &FromIni, &DataLength) )
		return(FromIni);						//Was in INI file
	else
		return(DefaultValue);
	}
#define GetProfileBool(KeyName, DefaultValue) (BOOL)GetProfileUlong(KeyName, (ULONG)DefaultValue)


/*===========================================================================*/
static void _Optlink GetProfileString( char  * KeyName,
																			 char  * Result,
																			 ULONG   LengthOfResultBuffer,
																			 char  * DefaultValue
																		 )
/*===========================================================================*/
	{
	PrfQueryProfileString(  HINI_USERPROFILE, APP_PROFILE_NAME,
													KeyName,
													DefaultValue,
													Result,
													LengthOfResultBuffer
											 );
	}


/*===========================================================================*/
static void _Optlink PutProfileUlong(char  * KeyName, ULONG Value)
/*===========================================================================*/
	{
	PrfWriteProfileData(HINI_USERPROFILE, APP_PROFILE_NAME, KeyName, &Value, sizeof(Value));
	}
#define PutProfileBool(KeyName, BooleanValue)   PutProfileUlong(KeyName, (ULONG)BooleanValue)


/*===========================================================================*/
static void _Optlink WriteProfileString(char  * KeyName, char  * Value)
/*===========================================================================*/
	{
	PrfWriteProfileString(HINI_USERPROFILE, APP_PROFILE_NAME, KeyName, Value);
	}



/*===========================================================================*/
static MRESULT _Optlink MenuWinSendMsg(ULONG msg, MPARAM mp1, MPARAM mp2)
/*===========================================================================*/
	{
	/*--- Send to context menu (subset of application menu) ------------------*/
	WinSendMsg(hApplicationMenu, msg, mp1, mp2);

	/*--- Send to application menu and return the return code ----------------*/
	return( WinSendMsg(hContextMenu, msg, mp1, mp2) );
	}


/*===========================================================================*/
static void _Optlink MenuWinCheckMenuItem(SHORT Id, BOOL Checked)
/*===========================================================================*/
	{
	/*--- Update both menus --------------------------------------------------*/
	WinCheckMenuItem(hContextMenu    , Id, Checked);
	WinCheckMenuItem(hApplicationMenu, Id, Checked);
	}


/*===========================================================================*/
static void _Optlink MenuWinEnableMenuItem(SHORT Id, BOOL Enabled)
/*===========================================================================*/
	{
	/*--- Update both menus --------------------------------------------------*/
	WinEnableMenuItem(hContextMenu,     Id, Enabled);
	WinEnableMenuItem(hApplicationMenu, Id, Enabled);
	}



/*===========================================================================*/
static void _Optlink DestroyHelpInstance(void)
/*===========================================================================*/
	{
	if ( hHelpInstance != NULLHANDLE )
		{
		WinDestroyHelpInstance(hHelpInstance);
		hHelpInstance = NULLHANDLE;
		}
	}



/*===========================================================================*/
static void _Optlink CreateHelpInstance(HWND Handle)
/*===========================================================================*/
	{
	HELPINIT    Help;
	char        HelpFileName[CCHMAXPATH];
#define HELP_SHORT_NAME  "PmPrintf.HLP"
	
	char    WindowTitle[2 * MAXNAMEL];

	/*--- Determine the name of the help file --------------------------------*/
	{
		PTIB    ThreadInfo;
		PPIB    ProcessInfo;
		char  * Ptr;
		ULONG   DosRc;

		/*--- Get program info -----------------------------------------------*/
		if ( (DosRc=DosGetInfoBlocks(&ThreadInfo, &ProcessInfo)) != 0 )
			{
			MsgBox(ATTR_ERROR,  "CAN'T GET PROCESS INFORMATION",
						 "This error may prevent you gaining access "
						 "to online help!"
						 "\n\n"
						 "The failing call was DosGetInfoBlocks() with a "
						 "return code of %lu."
						 ,DosRc
						);
			strcpy(HelpFileName, HELP_SHORT_NAME);
			}
		else
			{
			/*--- Get the full name of the EXE -------------------------------*/
			if ( (DosRc=DosQueryModuleName(ProcessInfo->pib_hmte, sizeof(HelpFileName), HelpFileName)) != 0 )
				{
				MsgBox(ATTR_ERROR,  "CAN'T GET MODULE NAME",
							 "This error may prevent you gaining access "
							 "to online help!"
							 "\n\n"
							 "The failing call was DosQueryModuleName() "
							 "with a return code of %lu."
							 ,DosRc
							);
				strcpy(HelpFileName, ProcessInfo->pib_pchcmd);	//Use name as per command line
				}

			/*--- Work out the name of the help file (Change .EXE to .HLP) ---*/
			strupr(HelpFileName);
			if ( (Ptr=strstr(HelpFileName, ".EXE")) != NULL )
				*Ptr = '\0';
			strcat(HelpFileName, ".HLP");
			if ( _access(HelpFileName, 0) != 0 )
				{
				MsgBox(ATTR_ERROR,  "HELP NOT IN PROGRAM DIRECTORY",
							 "We can't locate the file \"%s\"."
							 "\n\n"
							 "This error may prevent you gaining access "
							 "to online help!"
							 ,HelpFileName
							);
				strcpy(HelpFileName, HELP_SHORT_NAME);
				}
			}
	}

	/*--- Create the help instance -------------------------------------------*/
	memset(&Help, 0, sizeof(Help));
	Help.pszHelpLibraryName = HelpFileName;
	//Help.pszHelpWindowTitle = "PMPRINTF v"PGM_VERSION"";
	Help.pszHelpWindowTitle = "dummy";
	Help.pszTutorialName    = NULL;
	Help.phtHelpTable       = (PHELPTABLE)MAKELONG(IDHLPT_HELP_TABLE, 0xFFFF);
	if ( getenv("PMPRINTF_PANELID") == NULL )
		Help.fShowPanelId   = CMIC_HIDE_PANEL_ID;
	else
		Help.fShowPanelId		= CMIC_SHOW_PANEL_ID;		 //Show in debug mode
	Help.cb                 = sizeof(HELPINIT);
	
	sprintf (WindowTitle, "PmPrintf v%s", PGM_VERSION);
	if (strlen(WindowTitle)	> MAX_CLASS_NAME_LENGTH) WindowTitle[MAX_CLASS_NAME_LENGTH] = '\0'; 	// 20231230 cut title if too long
	Help.pszHelpWindowTitle = WindowTitle;
	
	hHelpInstance = WinCreateHelpInstance(hab, &Help);
	if ( hHelpInstance == 0L || Help.ulReturnCode )
		{
		/*--- Display message & disable help ---------------------------------*/
		MenuWinEnableMenuItem(IDM_HELP, FALSE);
		MsgBox(ATTR_ERROR,  "HELP UNAVAILABLE",
					 "We could not create a help instance."
					 "\n\n"
					 "We tried to load the help file \"%s\" and failed."
					 ,HelpFileName
					);
		DestroyHelpInstance();
		return;
		}

	/*--- Associate the help instance with window ----------------------------*/
	if ( ! WinAssociateHelpInstance(hHelpInstance, Handle) )
		if ( hHelpInstance == 0L || Help.ulReturnCode )
			{
			/*--- Display message & disable help ---------------------------------*/
			MenuWinEnableMenuItem(IDM_HELP, FALSE);
			MsgBox(ATTR_ERROR,  "HELP UNAVAILABLE", "We could not associate the help instance.");
			DestroyHelpInstance();
			return;
			}
	}



/*===========================================================================*/
static void DisplaySpecificHelpPanel(SHORT PanelId)
/*===========================================================================*/
	{
	if ( hHelpInstance != NULLHANDLE )
		WinSendMsg(hHelpInstance, HM_DISPLAY_HELP, MPFROMP(&PanelId), MPFROMSHORT(HM_RESOURCEID));
	}



/*===========================================================================*/
static void DisplayHelpContents(void)
/*===========================================================================*/
	{
	if ( hHelpInstance != NULLHANDLE )
		WinSendMsg(hHelpInstance, HM_HELP_CONTENTS, 0L, 0L);
	}


/*===========================================================================*/
static void SetNewLogFileMenuItemText(void)
/*===========================================================================*/
	{
	char    ShortFileName[30];
	char    NewText[100];

	/*--- Create a shorter version of the log file name ----------------------*/
	FmtShortenFileName(ShortFileName, NameOfLogFile, sizeof(ShortFileName)-1);

	/*--- Now set up the menu text -------------------------------------------*/
	sprintf(NewText, "Change Lo~gfile (%s)\tCtrl+G", ShortFileName);
	MenuWinSendMsg(MM_SETITEMTEXT, MPFROMSHORT(IDMI_CHANGE_LOGFILE_NAME), MPFROMP(NewText));
	}



/*===========================================================================*/
static void _Optlink RemoveLeadingAndTrailingWhiteSpace(char  * String)
/*===========================================================================*/
	{
	char    * DestinationPtr      = String;
	char    * StartTrailingSpaces = String;
	char      Character;

	/*--- Look for the start of the string ----------------------------------*/
	while ( *String == ' ' )
		String++;

	/*--- Copy the rest of the string (remember where last sig char is) -----*/
	while ( (Character=*String++) != '\0' )
		{
		/*--- Copy the character --------------------------------------------*/
		*DestinationPtr++ = Character;

		/*--- Remember where the last non white character is ----------------*/
		if ( Character != ' ' )
			StartTrailingSpaces = DestinationPtr;
		}

	/*--- Terminate the string (remove trailing whitespace) -----------------*/
	*StartTrailingSpaces = '\0';
	}



/*===========================================================================*/
static ULONG _Optlink QueryDlgItemTextTrimmed(HWND hwnd, ULONG Id, LONG MaxLength, char  *Buffer)
/*===========================================================================*/
	{
	ULONG   Rc;

	Buffer[0] = '\0';
	Rc = WinQueryDlgItemText(hwnd, Id, MaxLength, Buffer);
	RemoveLeadingAndTrailingWhiteSpace(Buffer);
	return(Rc);
	}

#define SetFieldLength(hwnd, id, length)            \
        WinSendMsg(WinWindowFromID(hwnd, id),       \
        EM_SETTEXTLIMIT, MPFROMSHORT(length), 0L)



/*--- The following are used by "QuestionDlgProc()" -------------------------*/
static char  * Title;
static char  * Answer;
static char  * DefaultAnswer;
static ULONG   AnswerBufferSize;
static BOOL    TrimAnswer;

/*===========================================================================*/
static MRESULT EXPENTRY QuestionDlgProc(HWND hwndDlg, ULONG msg, MPARAM mp1, MPARAM mp2)
/*===========================================================================*/
	{
	switch ( msg )																 /* Switch on message class   */
		{
		/*******************/
		case WM_INITDLG:
		/*******************/
			{
				/*--- Set the title for the question dialog ----------------------*/
				WinSetWindowText(hwndDlg, Title);

				/*--- Restrict what can be typed in ------------------------------*/
				SetFieldLength(hwndDlg, IDEF_ANSWER, AnswerBufferSize-1);

				/*--- Set up the current value for anser -------------------------*/
				WinSetDlgItemText(hwndDlg, IDEF_ANSWER, Answer);

				/*--- Is there a default action available? -----------------------*/
				if ( DefaultAnswer == NULL )
					WinEnableControl(hwndDlg, IDPB_DEFAULT, FALSE);		 //Disable control
				break;
			}
		/*******************/
		case WM_COMMAND:
		/*******************/
			{
				/*--- Which button was pressed? ----------------------------------*/
				switch ( SHORT1FROMMP(mp1) )
					{
					/*++++++++++++++++++*/
					case IDPB_DEFAULT:
						/*++++++++++++++++++*/
						{
							/*--- Overwrite current value with default ---------------*/
							WinSetDlgItemText(hwndDlg, IDEF_ANSWER, DefaultAnswer);
							return(FALSE);
						}
					/*++++++++++++++++++*/
					case DID_OK:
					/*++++++++++++++++++*/
						{
							if ( TrimAnswer )
								QueryDlgItemTextTrimmed(hwndDlg, IDEF_ANSWER, AnswerBufferSize-1, Answer);
							else
								WinQueryDlgItemText(hwndDlg, IDEF_ANSWER, AnswerBufferSize-1, Answer);
							WinDismissDlg(hwndDlg, TRUE);
							return(FALSE);
						}
					/*++++++++++++++++++*/
					case DID_CANCEL:
					/*++++++++++++++++++*/
						{
							WinDismissDlg(hwndDlg, FALSE); /* Return FALSE to caller */
							return(FALSE);
						}
					}
			}
		}

	/*--- Do default processing ----------------------------------------------*/
	return( WinDefDlgProc(hwndDlg, msg, mp1, mp2) );
	}



/*===========================================================================*/
static BOOL _Optlink AskQuestion(   char  * _Title,
																		char  * _Answer,
																		ULONG   _AnswerBufferSize,
																		char  * _DefaultAnswer,
																		BOOL    _TrimAnswer
																)
/*===========================================================================*/
	{
	BOOL    HaveAnswer;

	/*--- Remove the status message ------------------------------------------*/
	StatusBarClear();

	/*--- Set up the passed info ---------------------------------------------*/
	Title             = _Title;
	Answer            = _Answer;
	DefaultAnswer     = _DefaultAnswer;
	AnswerBufferSize  = _AnswerBufferSize;

	/*--- Ask the user for the answer ----------------------------------------*/
	HaveAnswer = WinDlgBox( HWND_DESKTOP,
													HWND_DESKTOP,
													QuestionDlgProc,
													0L,
													ID_QUESTION,
													NULL
												);

	/*--- Tell user if cancelled or not --------------------------------------*/
	return(HaveAnswer);
	}



/*===========================================================================*/
static SHORT _Optlink NumberOfSelectedItemsInListBox(SHORT * FirstSelected, SHORT * LastSelected)
/*===========================================================================*/
	{
	SHORT   Counter       = 0;
	SHORT   LastLocation  = LIT_NONE;
	SHORT   Location;

	/*--- Look for the 1st selected item -------------------------------------*/
	Location = (SHORT)WinSendMsg(hListBox, LM_QUERYSELECTION, MPFROMSHORT(LIT_FIRST), 0);
	if ( FirstSelected != NULL )
		*FirstSelected = Location;			//May be LIT_NONE!

	/*--- Work though all selected lines -------------------------------------*/
	while ( Location != LIT_NONE )
		{
		/*--- Keep count! ----------------------------------------------------*/
		Counter++;

		/*--- Look for the next item -----------------------------------------*/
		LastLocation = Location;
		Location = (SHORT)WinSendMsg(hListBox, LM_QUERYSELECTION, MPFROMSHORT(Location), 0);
		}

	/*--- Return to caller ---------------------------------------------------*/
	if ( LastSelected != NULL )
		*LastSelected = LastLocation;	 //May be LIT_NONE!
	return(Counter);									 //Return count of selected lines
	}



/*===========================================================================*/
static void _Optlink DeselectAllLinesInListBox(void)
/*===========================================================================*/
	{
	SHORT   Location;

	/*--- Work though all selected lines -------------------------------------*/
	Location = (SHORT)WinSendMsg(hListBox, LM_QUERYSELECTION, MPFROMSHORT(LIT_FIRST), 0);
	while ( Location != LIT_NONE )
		{
		/*--- Unselect this item ---------------------------------------------*/
		WinSendMsg(hListBox, LM_SELECTITEM, MPFROMSHORT(Location), MPFROMSHORT(FALSE));

		/*--- Look for the next item -----------------------------------------*/
		Location = (SHORT)WinSendMsg(hListBox, LM_QUERYSELECTION, MPFROMSHORT(Location), 0);
		}
	}


/*===========================================================================*/
static void _Optlink UpdateMaxLinesInListBoxMenuText(void)
/*===========================================================================*/
	{
	char    Text[50];

	sprintf(Text, "~Restrict listbox to ? lines (now %d)", MaxLinesKeptInWindow);
	MenuWinSendMsg(MM_SETITEMTEXT,  MPFROMSHORT(IDMI_MAX_LINES_IN_LISTBOX), MPFROMP(Text));
	}



/*===========================================================================*/
static void _Optlink IfNotVisiblePutInMiddleOfListBox(SHORT Item)
/*===========================================================================*/
	{
	SHORT  FirstItem = (SHORT)WinSendMsg(hListBox, LM_QUERYTOPINDEX, 0, 0);
	SHORT  LastItem  = FirstItem + LinesVisibleInListbox - 1;

	/*--- Is the item visible? -----------------------------------------------*/
	if ( (Item >= FirstItem) && (Item <= LastItem) )
		return;											//Item can be seen

	/*--- Item is not visible so move it to center ---------------------------*/
	FirstItem = Item - (LinesVisibleInListbox / 2);
	if ( FirstItem < 0 )
		FirstItem = 0;
	WinSendMsg(hListBox, LM_SETTOPINDEX, MPFROMSHORT(FirstItem), 0);
	}



/*===========================================================================*/
static void _Optlink HandleFindResult(SHORT FoundItem)
/*===========================================================================*/
	{
	if ( FoundItem == LIT_NONE )
		StatusBarMessage(BEEP_ERROR, "Could not find the string \"%s\"", SearchString);
	else
		{
		/*--- Found a match! -------------------------------------------------*/
		LastMatchLocation = FoundItem;
		WinSendMsg(hListBox, LM_SELECTITEM, MPFROMSHORT(FoundItem), MPFROMSHORT(TRUE));
		IfNotVisiblePutInMiddleOfListBox(FoundItem);
		StatusBarMessage(FALSE, "String found on line %hd", FoundItem+1);
		}
	}



/*===========================================================================*/
static SHORT _Optlink Find(     SHORT From,
																SHORT To
#define TO_END  0x7FFF
													)
/*===========================================================================*/
	{
	SHORT   Item;
	char    Line[MAX_LINE_LENGTH];
	SHORT   Increment;
	SHORT   StopWhenHere;

	/*--- Need to work out where to end? -------------------------------------*/
	if ( To == TO_END )
		To = WinQueryLboxCount(hListBox);

	/*--- Work out whether searching forwards or backwards -------------------*/
	if ( To >= From )
		Increment = 1;								//Search Forwards
	else
		Increment	= -1;								//Search Backwards
	StopWhenHere  = To + Increment;

	/*--- Loop through looking for next match --------------------------------*/
	for ( Item=From; Item != StopWhenHere; Item += Increment )
		{
		/*--- Get the text of the item ---------------------------------------*/
		Line[0] = '\0';
		WinQueryLboxItemText(hListBox, Item, Line, sizeof(Line));

		/*--- See if we have a match -----------------------------------------*/
		if ( CaseSensitiveSearches )
			{
			/*--- Do case sensitive search -----------------------------------*/
			if ( strstr(Line, SearchString) != NULL )
				return(Item);								//Found a match!
			}
		else
			{
			/*--- Do case insensitive search ---------------------------------*/
			if ( strIstr(Line, SearchString) != NULL )
				return(Item);								//Found a match!
			}
		}
	return(LIT_NONE);										//Match was not found
	}




/*===========================================================================*/
static BOOL _Optlink FindPrevious(void)
/*===========================================================================*/
	{
	SHORT   Item;
	SHORT   FoundItem = LIT_NONE;

	/*--- Find out where to start search from --------------------------------*/
	StatusBarString(FALSE, "Searching Backwards....");
	if ( LastMatchLocation == LIT_NONE )
		{
		HandleFindResult(LIT_NONE);
		return(LIT_NONE);
		}
	Item = LastMatchLocation - 1;						//Start from previous line

	/*--- Make sure all lines are unhighlighted ------------------------------*/
	DeselectAllLinesInListBox();

	/*--- Look for the string ------------------------------------------------*/
	if ( Item != -1 )
		FoundItem = Find(Item, 0);
	if ( FoundItem == LIT_NONE )
		{
		SHORT LastItem = WinQueryLboxCount(hListBox);

		/*--- Not found, do we need to search first part? --------------------*/
		if ( Item != LastItem )
			FoundItem = Find(LastItem, Item+1);		//Search back (wrap around)
		}

	/*--- Process the find results -------------------------------------------*/
	HandleFindResult(FoundItem);
	return(FoundItem);
	}



/*===========================================================================*/
static BOOL _Optlink FindNext(void)
/*===========================================================================*/
	{
	SHORT   Item;
	SHORT   FoundItem;

	/*--- Find out where to start search from --------------------------------*/
	StatusBarString(FALSE, "Searching Forwards....");
	if ( LastMatchLocation != LIT_NONE )
		Item = LastMatchLocation + 1;						//Start from following line
	else
		{
		NumberOfSelectedItemsInListBox(&Item, NULL);
		if ( Item == LIT_NONE )
			Item = 0;														 //Start with 1st item in listbox
		else
			Item++;
		}

	/*--- Make sure all lines are unhighlighted ------------------------------*/
	DeselectAllLinesInListBox();

	/*--- Look for the string ------------------------------------------------*/
	if ( (FoundItem=Find(Item, TO_END)) == LIT_NONE )
		{
		/*--- Not found, do we need to search first part? --------------------*/
		if ( Item != 0 )
			FoundItem = Find(0, Item-1);					//Search front (wrap around)
		}

	/*--- Process the find results -------------------------------------------*/
	HandleFindResult(FoundItem);
	return(FoundItem);
	}





/*===========================================================================*/
static BOOL _Optlink ChooseFileForLoggingTo(void)
/*===========================================================================*/
	{
	FILEDLG FileDialog;

	/*--- Set up the File Dialog structure -----------------------------------*/
	memset(&FileDialog, 0, sizeof(FileDialog));
	strcpy(FileDialog.szFullFile, NameOfLogFile);
	FileDialog.cbSize      = sizeof(FileDialog);
	FileDialog.pszTitle    = "Please select a new name for the log file";
	FileDialog.pszOKButton = "~Set Log File's Name";
	FileDialog.fl          = FDS_CENTER
													 | FDS_SAVEAS_DIALOG
													 | FDS_ENABLEFILELB;

	/*--- Get the name of the file -------------------------------------------*/
	if ( ! WinFileDlg(HWND_DESKTOP, hFrame, &FileDialog) )
		return(FALSE);
	if ( FileDialog.lReturn != DID_OK )
		return(FALSE);

	/*--- See if we actually have a new filename -----------------------------*/
	if ( stricmp(NameOfLogFile, FileDialog.szFullFile) == 0 )
		return(FALSE);											//Filename unchanged
	else
		{
		/*--- The user did choose a new filename -----------------------------*/
		strcpy(NameOfLogFile, FileDialog.szFullFile);
		strupr(NameOfLogFile);
		return(TRUE);
		}
	}




/*===========================================================================*/
static VOID OpenLogFileAndTurnOnLogging(void)
/*===========================================================================*/
	{
	ULONG   DosRc;
	ULONG   OpenAction;
	ULONG   FilePointer;

	/*--- Open the log file --------------------------------------------------*/
	DosRc = DosOpen(    NameOfLogFile, &LogfileHandle, &OpenAction, 0L,
											FILE_NORMAL,
											FILE_OPEN | FILE_CREATE,
											OPEN_ACCESS_READWRITE
											| OPEN_SHARE_DENYWRITE
											| OPEN_FLAGS_SEQUENTIAL,
											0L
								 );
	if ( DosRc != 0 )
		{
		/*--- Open Failed ----------------------------------------------------*/
		MsgBox(ATTR_ERROR, "COULD NOT OPEN LOG FILE",
					 "We could not open the log file (%s), a return code of "
					 "%lu was received."
					 "\n\n"
					 "If you have the log file open in an editor or have "
					 "another instance of this program running (and logging to "
					 "the same file) then thats likely to be the cause."
					 ,NameOfLogFile
					 ,DosRc
					);
		}
	else
		{
		/*--- Turn Logging on ------------------------------------------------*/
		LoggingToFile = TRUE;
		PutProfileBool(KEY_LOGGING_RECEIVED_LINES,       LoggingToFile);
		MenuWinCheckMenuItem(IDMI_LOG_NEW_LISTBOX_LINES, LoggingToFile);

		/*--- Move to EOF (remove EOF char if there - from editors) ----------*/
		if ( DosSetFilePtr(LogfileHandle, -1L, FILE_END, &FilePointer) == 0 )
			{
			char  OneByte;
			ULONG BytesRead;

			/*--- Now 1 byte off EOF, read this character --------------------*/
			if ( DosRead(LogfileHandle, &OneByte, 1U, &BytesRead) == 0 )
				{
				/*--- We have 1 character, and are at EOF, is char EOF char? ---*/
				if ( OneByte == '\x1a' )
					{
					/*--- Last byte is EOF char, go back one so next write will remove it ---*/
					DosSetFilePtr(LogfileHandle, -1L, FILE_END, &FilePointer);
					}
				}
			}
		}
	return;
	}




/*===========================================================================*/
static VOID CloseLogfileAndStopLogging(void)
/*===========================================================================*/
	{
	/*--- Close the file -----------------------------------------------------*/
	DosClose(LogfileHandle);

	/*--- Update display etc -------------------------------------------------*/
	LoggingToFile = FALSE;
	PutProfileBool(KEY_LOGGING_RECEIVED_LINES,       LoggingToFile);
	MenuWinCheckMenuItem(IDMI_LOG_NEW_LISTBOX_LINES, LoggingToFile);
	}


/*===========================================================================*/
static void _System FatalError(int ExitRc, char  * Format, ...)
/*===========================================================================*/
	{
	va_list  Args;

	/*--- Beep ---------------------------------------------------------------*/
	DosBeep(1000,80);

	/*--- Display the error message in a PM popup ----------------------------*/
	va_start(Args, Format);
	MsgBoxVa(ATTR_ERROR, "FATAL ERROR", Format, Args);
	va_end(Args);

	/*--- Exit to dos --------------------------------------------------------*/
	exit(ExitRc);
	}



/*===========================================================================*/
static void _Optlink HandleMsgFromQueue(char * Msg, time_t MsgTimeStamp, BOOL Beep, ULONG NumberMsgsInQueue)
/*===========================================================================*/
	{
	/*--- Update global varible of number of messages in Queue ---------------*/
	NumberMessagesInNamedQueue = NumberMsgsInQueue;

	/*--- We may need to beep ------------------------------------------------*/
	if ( Beep )
		{
		DosBeep(1000,80);
		DosBeep(1300,80);
		DosBeep(1000,80);
		}

	/*--- Wait until not Paused! ---------------------------------------------*/
	DosWaitEventSem(hSemPaused, (unsigned)SEM_INDEFINITE_WAIT);

	/*--- Keep posting until successful --------------------------------------*/
	/* while ( ! WinPostMsg(hClient, UM_GOTLINE, MPFROMP(Msg), MPFROMLONG(MsgTimeStamp)));
		DosSleep(500L);
	*/
	// 20231230 AB change from WinPosMsg to WinSendMsg fixes lost messages
	// 	AND consumes much less lower shared memory AND is much faster at least on SMP systems
	// no clue why it's possible to lose message with WinPostMsg in the above while loop
	WinSendMsg(hClient, UM_GOTLINE, MPFROMP(Msg), MPFROMLONG(MsgTimeStamp));
	}




/*===========================================================================*/
static void _Optlink QueueThread(void * UnusedThreadArgument)
						/*                                                                           */
						/* When input arrives, it is read and sent via an UM_GOTLINE message         */
						/* to the main WinProc.  Two semaphores are used to control pausing          */
						/* and sharing of memeory.                                                   */
						/*                                                                           */
						/* Protocol:                                                                 */
						/*   PRINTF writes its data to the named queue using the following           */
						/*   protocol:                                                               */
						/*                                                                           */
						/*     Request -- PID ID in the first LONG, and the timestamp in             */
						/*                the second (format as above).                              */
						/*     Length  -- The length of the data, including terminator.              */
						/*                A negative length indicates a BELL in the data.            */
						/*     Address -- Address of the data.                                       */
/*===========================================================================*/
	{
	EXCEPTIONREGISTRATIONRECORD exRegRec;
	HAB           hab2;
	HMQ           hmp2;
	ULONG         DosRc;
	REQUESTDATA   Request;
	ULONG         MessageSize;
	char        * ReceivedPtr;
	BYTE          MessagePriority;
	time_t        MsgTimeStamp;
	ULONG         NumberMessages;

	LoadExceptq(&exRegRec, NULL, NULL);
	/*--- Kick start PM ------------------------------------------------------*/
	hab2 = WinInitialize(0);
	hmp2 = WinCreateMsgQueue(hab2,0);
	WinCancelShutdown(hmp2,TRUE);

	/*--- For now lets hope that queue is OK! --------------------------------*/
	WinPostMsg(hClient, UM_QUEUE_CREATED, NULL, NULL);

	/*--- Create The Queue (and process the messages -------------------------*/
	CreatePmprintfQueue(FullQueueName, HandleMsgFromQueue, FatalError);

	UninstallExceptq(&exRegRec);
	}


/*===========================================================================*/
static void TogglePausedState(void)
/*===========================================================================*/
	{
	ULONG  PostedCount;
	UINT   Highlight;

	/*--- Changed State ------------------------------------------------------*/
	QueuePaused = ! QueuePaused;

	/*--- What state are we going into ---------------------------------------*/
	if ( QueuePaused )
		{
		/*--- Stop messages being processed off the Queue --------------------*/
		DosResetEventSem(hSemPaused, &PostedCount);

		/*--- Change Text of this control to "Restart" -----------------------*/
		MenuWinSendMsg(MM_SETITEMTEXT,  MPFROMSHORT(IDMI_PAUSE_RESUME), MPFROMP("~Restart"));

		/*--- Determine the required highlighting ----------------------------*/
		Highlight = MIA_FRAMED | MIA_HILITED;
		}
	else
		{
		/*--- Change Text of this control to "Pause" -------------------------*/
		MenuWinSendMsg(MM_SETITEMTEXT, MPFROMSHORT(IDMI_PAUSE_RESUME), MPFROMP("~Halt"));

		/*--- Allow messages to be processed off the Queue -------------------*/
		DosPostEventSem(hSemPaused);

		/*--- Determine the required highlighting ----------------------------*/
		Highlight = 0;
		}

	/*--- Set Highlight state of control (only on Action bar!) ---------------*/
	WinSendMsg( hApplicationMenu,
							MM_SETITEMATTR,
							MPFROM2SHORT(IDMI_PAUSE_RESUME, TRUE),
							MPFROM2SHORT(MIA_FRAMED | MIA_HILITED, Highlight)
						);
	}




/*===========================================================================*/
static USHORT _Optlink AddLineToListbox(char  * StringToAdd, time_t MsgTimeStamp, SHORT Position)
/*===========================================================================*/
	{
	char  * MsgToDisplay;
	char    Buffer[70+MAX_LINE_LENGTH+1];
	SHORT   PositionOfInsertedItem;
	APIRET  rc = NO_ERROR;
	ULONG ulMemSize = 1;
	ULONG ulMemFlags;
	char *HighMemBuffer;
	char cBuf[4];
	int i;
	char buf[1024];
	char szMessage[] = "PMPrintf ERROR: got invalid string, invalid memory ERROR at";

	rc = DosQueryMem(StringToAdd, &ulMemSize, &ulMemFlags);		// 20181103 AB check for valid string
	if ( rc != NO_ERROR || !(ulMemFlags & PAG_READ) )
		{	// invalid mem
		sprintf(buf, "%s 0x%lX (%s, %s, %d)", szMessage,  StringToAdd, __FILE__, __FUNCTION__, __LINE__);
		HighMemBuffer = buf;
		}
	else
		{	 // only check pointer to highmem when pointer in lower shared mem is valid
		memcpy(&HighMemBuffer, StringToAdd, sizeof(char *) );
		rc = DosQueryMem(HighMemBuffer, &ulMemSize, &ulMemFlags);		// 20230702 AB check for valid mem pointer
		if ( rc != NO_ERROR || !(ulMemFlags & PAG_READ) )
			{	// invalid mem
			sprintf(buf, "%s 0x%lX (%s, %s, %d)", "PmPrintf ERROR: invalid HighMem buffer",  HighMemBuffer, __FILE__, __FUNCTION__, __LINE__);
			HighMemBuffer = buf;
			// try to concenate as it would be old style, none HighMem message
			//vsprintf(buf + strlen(buf), StringToAdd, args);
			strcat(buf + strlen(buf), StringToAdd);
			}
	
		/*--- Free the messages memory. In case of HighMem this is only the pointer -------*/
		rc = DosFreeMem(StringToAdd);
		if (rc)
			{
			sprintf(buf + strlen(buf), " ERROR DosFreeMem rc=%d (%s %s)", rc, __FILE__, __LINE__);
			}

		}

	/*--- If message time not supplied, obtain current time ------------------*/
	if ( MsgTimeStamp == 0 )
		time(&MsgTimeStamp);

	/*--- Keep as many lines as wanted, or as many as can be seen in window --*/
	if ( LinesInListbox >= MaxLinesKeptInWindow && LinesInListbox > LinesVisibleInListbox )
		{
		/*--- Need to remove a line ------------------------------------------*/
		WinSendMsg(hListBox, LM_DELETEITEM, MPFROMSHORT(0), HighMemBuffer);
		LinesInListbox--;
		}

	/*--- See if we wish to attach the time to the message -------------------*/
	if ( (! AddTimeToDisplayedLines) || (*HighMemBuffer == '\0') )
		MsgToDisplay = HighMemBuffer;			//Use as is
	else
		{
		/*--- Add the time to the message ------------------------------------*/
		strftime(Buffer, sizeof(Buffer), LogTimeFormat, localtime(&MsgTimeStamp));
		strncat(Buffer, HighMemBuffer, 70+MAX_LINE_LENGTH - strlen(Buffer));
		MsgToDisplay = Buffer;					//Use string with time prefix
		}

	/*--- Add the line to the listbox ----------------------------------------*/
	PositionOfInsertedItem = (SHORT)WinSendMsg(hListBox, LM_INSERTITEM, MPFROMSHORT(Position), MsgToDisplay);

	/*--- Keep count of lines (check if past end of visible region) ----------*/
	if ( ++LinesInListbox > LinesVisibleInListbox )
		{
		/*--- If message is being added to end (ensure we can see it) --------*/
		if ( Position == LIT_END )
			{
			/*--- Scroll if nothing waiting to be added --------------------------*/
			if ( NumberMessagesInNamedQueue == 0 )
				WinSendMsg(hListBox, LM_SETTOPINDEX, MPFROMSHORT(LinesInListbox-LinesVisibleInListbox), NULL);
			}
		}

	/*--- If beep on message arrival set then beep! --------------------------*/
	if ( BeepOnMessageArrival && NumberMessagesInNamedQueue == 0 )
		DosBeep(3000,80);

	/*--- If we were asked to log the info -----------------------------------*/
	if ( LoggingToFile )
		{
		ULONG    CharsInMessage;
		ULONG    TimeCharsWritten;
		ULONG    MsgCharsWritten;
		ULONG    CrlFCharsWritten;
		ULONG    DosRc;
#define  CRLF  "\r\n"

		/*--- If we want a timestamp (and have one!) then log it -------------*/
		if ( LogTimeWithMsgToFile )
			{
			/*--- Get the time from the message & format it ------------------*/
			strftime(Buffer, sizeof(Buffer), LogTimeFormat, localtime(&MsgTimeStamp));

			/*--- Write the time to the file ---------------------------------*/
			DosWrite(LogfileHandle, Buffer, strlen(Buffer), &TimeCharsWritten);
			}

		/*--- Now output the received message --------------------------------*/
		CharsInMessage  = strlen(HighMemBuffer);
		DosRc = DosWrite(LogfileHandle, HighMemBuffer, CharsInMessage, &MsgCharsWritten);

		// HighMem buffer is not needed anymore, free it up
		rc = DosFreeMem(HighMemBuffer);
    if ( rc != 0 ) {
			sprintf(Buffer, "*** ERROR DosFreeMem rc=%d (%s %s)", rc, __FILE__, __LINE__);
			WinSendMsg(hListBox, LM_INSERTITEM, (MPARAM)LIT_END, Buffer);
			}

		/*--- If no error so far output a CR & LF ----------------------------*/
		if ( DosRc == 0 )
			DosRc = DosWrite(LogfileHandle, CRLF, sizeof(CRLF)-1, &CrlFCharsWritten);

		/*--- Report any write errors ----------------------------------------*/
		if ( DosRc != 0 || (long)MsgCharsWritten != CharsInMessage || (long)CrlFCharsWritten != sizeof(CRLF)-1 ) {
//			WinPostMsg(hClient, UM_ERROR_WRITING_TO_LOG_FILE, MPFROMSHORT((short)DosRc), NULL);
			sprintf(Buffer, "*** ERROR UM_ERROR_WRITING_TO_LOG_FILE rc=%d (%s %s)", DosRc, __FILE__, __LINE__);
			WinSendMsg(hListBox, LM_INSERTITEM, (MPARAM)LIT_END, Buffer);
			WinSendMsg(hClient, UM_ERROR_WRITING_TO_LOG_FILE, MPFROMSHORT((short)DosRc), NULL);		// 20231231 try to fix lost messages
			}

		/*--- Flush the I/O if  no queued messages ---------------------------*/
		if ( NumberMessagesInNamedQueue == 0 )
			DosResetBuffer(LogfileHandle);
		}

	/*--- Tell the caller where the item was positioned ----------------------*/
	return(PositionOfInsertedItem);
	}



/*===========================================================================*/
static MRESULT EXPENTRY SubclassListBoxProcedure(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
/*===========================================================================*/
	{
	switch ( msg )
		{
		/*++++++++++++++++++++++*/
		case WM_CONTEXTMENU:
		/*++++++++++++++++++++++*/
			{
				POINTL  MenuPositionBLC;

				/*--- Hide the application menu if its up ------------------------*/
				WinSendMsg(hApplicationMenu, MM_ENDMENUMODE, MPFROMSHORT(TRUE), 0);

				/*--- Work out where to put the bottom left part of menu (if possible) ---*/
				WinQueryPointerPos(HWND_DESKTOP, &MenuPositionBLC);
				WinMapWindowPoints(HWND_DESKTOP, hwnd, &MenuPositionBLC, 1);

				/*--- Display the menu -------------------------------------------*/
				WinPopupMenu(   hFrame, hClient, hContextMenu,
												MenuPositionBLC.x, MenuPositionBLC.y,
												IDM_ACTIONS,
												PU_NONE
												| PU_HCONSTRAIN
												| PU_VCONSTRAIN
												| PU_MOUSEBUTTON1
												| PU_MOUSEBUTTON2
												| PU_MOUSEBUTTON3
												| PU_KEYBOARD
										);
				return(     (     MRESULT)TRUE);
			}
		/*++++++++++++++++++++++*/
		case WM_BUTTON1DOWN:
		case WM_BUTTON2DOWN:
		case WM_BUTTON3DOWN:
		/*++++++++++++++++++++++*/
			{
				/*--- Remove both menus ------------------------------------------*/
				WinSendMsg(hContextMenu,     MM_ENDMENUMODE, MPFROMSHORT(TRUE), 0);
				WinSendMsg(hApplicationMenu, MM_ENDMENUMODE, MPFROMSHORT(TRUE), 0);

				/*--- Clear the status bar ---------------------------------------*/
				StatusBarClear();
				break;
			}
		}

	/*--- Do default processing ----------------------------------------------*/
	return( ListBoxFunction(hwnd, msg, mp1, mp2) );
	}



/*===========================================================================*/
static MRESULT EXPENTRY PmprintfWindowProcedure(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
/*===========================================================================*/
	{
	SHORT   ControlId;
	CHAR    stuff[CCHMAXPATH+201];	/* buffer */
	MRESULT mr;												/* receives message results */
	POINTL  xy;												 /* an x-y point */
	ULONG   urc;
	ULONG   lrc;
	int     len, item;						/* .. */

	switch ( msg )
		{
		/**************************************/
		case WM_CREATE:
		/**************************************/
			{
				/*--- Window not yet created, finish later! ----------------------*/
				WinPostMsg(hwnd, UM_CREATED, NULL, NULL);
				return((MRESULT)FALSE);
			}
		/*********************/
		case UM_CREATED:
		/*********************/
			{
				SWCNTRL SwitchListCntl;
				//	char    WindowTitle[MAXNAMEL + 1];
				char    WindowTitle[2 * MAXNAMEL];
				RECTL   rcl;

				/*--- Record handle of the Action Bar menu -----------------------*/
				hApplicationMenu = WinWindowFromID(hFrame, FID_MENU);

				/*--- Load the context menu --------------------------------------*/
				hContextMenu     = WinLoadMenu(hwnd, NULLHANDLE, ID_CONTEXT_MENU);

				/*--- Set up the windows title -----------------------------------*/
				if ( stricmp(CmdLineOptions.ShortQueueName, DEFAULT_PRINTF_QUEUE_NAME) == 0 )
					sprintf(WindowTitle,"PmPrintf32 - %s", PGM_VERSION);
				else
					sprintf(WindowTitle,"PmPrintf32 - %s", CmdLineOptions.ShortQueueName);
				if (strlen(WindowTitle)	>= MAXNAMEL) WindowTitle[MAXNAMEL-1] = '\0'; 	// 20231230 cut title if too long
				WinSetWindowText(hFrame, WindowTitle);

				/*--- Set up the Switch list entry -------------------------------*/
				memset(&SwitchListCntl, 0, sizeof(SwitchListCntl));
				SwitchListCntl.hwnd = hFrame;
				strcpy(SwitchListCntl.szSwtitle, "PmPrintf - By Dennis Bareis & AB");
				WinAddSwitchEntry(&SwitchListCntl);

				/*--- Read last used values from profile -------------------------*/
				AddTimeToDisplayedLines = GetProfileBool(KEY_TIME_IN_LISTBOX,            AddTimeToDisplayedLines);
				LogTimeWithMsgToFile    = GetProfileBool(KEY_ADD_TIME_TO_LOGGED_LINES,   LogTimeWithMsgToFile);
				RestorePresParms        = GetProfileBool(KEY_RESTORE_PRESPARMS,          RestorePresParms);
				SavePresParms           = GetProfileBool(KEY_SAVE_PRESPARMS,             SavePresParms);
				CaseSensitiveSearches   = GetProfileBool(KEY_CASE_SENSITIVE_SEARCHES,    CaseSensitiveSearches);
				MaxLinesKeptInWindow    = GetProfileUlong(KEY_MAX_LINES_IN_LISTBOX, MaxLinesKeptInWindow);

				/*--- Update checked items etc -----------------------------------*/
				MenuWinCheckMenuItem(IDMI_ADD_TIME_TO_LISTBOX_ENTRIES,   AddTimeToDisplayedLines);
				MenuWinCheckMenuItem(IDMI_ADD_TIMESTAMP_TO_LOGGED_LINES, LogTimeWithMsgToFile);
				MenuWinCheckMenuItem(IDMI_RESTORE_PRESPARMS,             RestorePresParms);
				MenuWinCheckMenuItem(IDMI_SAVE_PRESPARMS,                SavePresParms);
				MenuWinCheckMenuItem(IDMI_CASE_SENSITIVE_SEARCHES,       CaseSensitiveSearches);
				UpdateMaxLinesInListBoxMenuText();

				/*--- Get the name of the log file (update menu) -----------------*/
				GetProfileString(KEY_NAME_OF_LOG_FILE, NameOfLogFile, sizeof(NameOfLogFile), DEFAULT_LOG_NAME);
				SetNewLogFileMenuItemText();

				/*--- Start logging? ---------------------------------------------*/
				if ( GetProfileBool(KEY_LOGGING_RECEIVED_LINES, FALSE) )
					OpenLogFileAndTurnOnLogging();

				/*--- Get some other values --------------------------------------*/
				GetProfileString(KEY_TIME_FORMAT,   LogTimeFormat, sizeof(LogTimeFormat), DEFAULT_TIME_FMT);
				GetProfileString(KEY_SEARCH_STRING, SearchString,  sizeof(SearchString),  "");

				/*--- Create the listbox -----------------------------------------*/
				WinQueryWindowRect(hwnd, &rcl);		 /* need startup size */
				hListBox = WinCreateWindow( hwnd,
																		WC_LISTBOX,
																		"Paper",
																		WS_VISIBLE | LS_HORZSCROLL | LS_EXTENDEDSEL,
																		0, 0,
																		rcl.xRight, rcl.yTop,
																		hwnd, HWND_TOP,
																		IDLB_LISTBOX,
																		NULL, NULL
																	);

				/*--- Set the default Listbox font -------------------------------*/
				WinSetPresParam(hListBox, PP_FONTNAMESIZE, sizeof(DEFAULT_LISTBOX_FONT), DEFAULT_LISTBOX_FONT);

				/*--- Restore the windows position, fonts & colors ---------------*/
				RestorePositionAndPresentationParameters(hFrame);

				/*--- Make sure listbox has focus --------------------------------*/
				WinSetFocus(HWND_DESKTOP, hListBox);

				/*--- Subclass the Listbox window Procedure ----------------------*/
				ListBoxFunction = WinSubclassWindow(hListBox, SubclassListBoxProcedure);

				/*--- Initialize help --------------------------------------------*/
				CreateHelpInstance(hFrame);

				/*--- Begin the thread which will read the queue -----------------*/
				_beginthread(QueueThread, NULL, 20480, NULL);
				return((MRESULT)FALSE);
			}
		/*********************/
		case UM_SIZE:				/* Adjust frame window please */
		/*********************/
			{
				int x, y, cx, cy;

				x =SHORT1FROMMP(mp1); y =SHORT2FROMMP(mp1);
				cx=SHORT1FROMMP(mp2); cy=SHORT2FROMMP(mp2);
				WinSetWindowPos( hFrame, 0, x, y, cx, cy,	 /* causes WM_SIZE */
												 (unsigned)SWP_SIZE|(unsigned)SWP_MOVE
											 );
				return((MRESULT)TRUE);
			}
		/*********************/
		case WM_SIZE:				/* Resizing client window */
		/*********************/
			{
				int xr, yt, ly, adjust;
				SWP swp;											 /* window position structure */

				xr=SHORT1FROMMP(mp2); yt=SHORT2FROMMP(mp2);
				/* Set list box to same size */
				WinSetWindowPos(hListBox, 0, 0, 0, xr, yt, (unsigned)SWP_SIZE|(unsigned)SWP_NOREDRAW);

				/* The list box resizes to click to integer number of lines... */
				WinQueryWindowPos(hListBox, &swp);			 /* get actual size */
				ly=swp.cy;															 /* listbox height */
				adjust=yt-ly+2;													 /* size of the gap */
				/* (note fudge factor) */
				/* slight danger here if resize not stable :-( */
				if ( xr>0 && yt>0 )
					{						 /* not start-up or minimized */
					if ( adjust!=0 )
						{	/* Squish the frame to close the gap */
						WinQueryWindowPos(hFrame, &swp);		/* get actuals of frame */
						WinPostMsg(hwnd, UM_SIZE,
											 MPFROM2SHORT((SHORT)swp.x, (SHORT)swp.y +adjust),
											 MPFROM2SHORT((SHORT)swp.cx,(SHORT)swp.cy-adjust));
						}
					}	/* non-zero size */

				/* Convert pels into dialog units -- so we can see how many lines */
				xy.x=xr; xy.y=ly;
				WinMapDlgPoints(hwnd, &xy, 1, FALSE);		 /* -> Dlg units */
				/* 8 du/line; the horizontal scroll bar is 1.x lines: */
				LinesVisibleInListbox = (xy.y / 8) - 1;
				return(MRESULT)TRUE;
			}

		/*********************/
		case WM_COMMAND:		 /* Asynchronous controls (Buttons) */
		/*********************/
			{
				/*--- Now what was requested? ------------------------------------*/
				switch ( SHORT1FROMMP(mp1) )
					{
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_CHANGE_LOGFILE_NAME:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							strcpy(stuff, NameOfLogFile);		//Remember original filename
							if ( ChooseFileForLoggingTo() )
								{
								/*--- Filename modified, save new name to INI file ---*/
								WriteProfileString(KEY_NAME_OF_LOG_FILE, NameOfLogFile);

								/*--- Change menu text -------------------------------*/
								SetNewLogFileMenuItemText();

								/*--- If logging, close old file & open new ----------*/
								if ( LoggingToFile )
									{
									CloseLogfileAndStopLogging();
									OpenLogFileAndTurnOnLogging();
									}
								}
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_WRITE_LISTBOX_TO_LOGFILE:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							int savelog;										/* saves LoggingToFile setting */
							int savepaused;									/* saves paused setting */
							PSZ    linebuff[MAX_LINE_LENGTH+2];			/* text receptor */
							ULONG count, count2;					 /* counts of bytes written */
							int hadtrunc=FALSE;							/* a line was too long */
							int i;
							char Buffer[256];

							if ( LinesInListbox==0 ) break;						 /* nothing to do */

							savelog=LoggingToFile;
							savepaused=QueuePaused;							 /* save */

							/*--- Prevent interference -------------------------------*/
							MenuWinEnableMenuItem(IDM_ACTIONS,       FALSE);
							MenuWinEnableMenuItem(IDM_OPTIONS,       FALSE);
							MenuWinEnableMenuItem(IDMI_PAUSE_RESUME, FALSE);

							if ( !savepaused )
								TogglePausedState();		 /* pause if necessary */
							if ( !savelog )
								OpenLogFileAndTurnOnLogging();			 /* start (open) logging */
							if ( LoggingToFile )
								{
								/* Log is open and ready-to-write */
								for ( i=0; i<(short)LinesInListbox; i++ )
									{/* for each line */
									mr=WinSendMsg(hListBox, LM_QUERYITEMTEXT,
																MPFROM2SHORT((short)i, (short)(MAX_LINE_LENGTH+2)), MPFROMP(linebuff));
									len=SHORT1FROMMR(mr);
									if ( len>MAX_LINE_LENGTH )
										{
										linebuff[MAX_LINE_LENGTH]='\0';			/* truncate */
										hadtrunc=TRUE;
										}
									urc=DosWrite(LogfileHandle, linebuff, (ULONG)len, &count);
									if ( urc==0 )	urc=DosWrite(LogfileHandle, "\r\n", 2, &count2); /* CRLF */
									if ( urc!=0 || (long)count!=(long)len || (long)count2!=2 )
										{
										//WinPostMsg(hClient, UM_ERROR_WRITING_TO_LOG_FILE, MPFROMSHORT((short)urc), NULL);
										WinSendMsg(hClient, UM_ERROR_WRITING_TO_LOG_FILE, MPFROMSHORT((short)urc), NULL);		// 20231231 try to fix lost messages
										sprintf(Buffer, "*** ERROR UM_ERROR_WRITING_TO_LOG_FILE rc=%d (%s %s)", urc, __FILE__, __LINE__);
										WinSendMsg(hListBox, LM_INSERTITEM, (MPARAM)LIT_END, Buffer);
										break;
										}	/* from I-loop */
									}	/* i */
								if ( !savelog )	CloseLogfileAndStopLogging();					/* close if needed */
								}	/* opened ok */
							if ( !savepaused )
								TogglePausedState();		 /* and unpause if needed */

							/*--- Allow interference??? ------------------------------*/
							MenuWinEnableMenuItem(IDM_ACTIONS,       TRUE);
							MenuWinEnableMenuItem(IDM_OPTIONS,       TRUE);
							MenuWinEnableMenuItem(IDMI_PAUSE_RESUME, TRUE);

							if ( hadtrunc )
								// WinPostMsg(hClient, UM_FILETRUNC, NULL, NULL);
								WinSendMsg(hClient, UM_FILETRUNC, NULL, NULL);		// 20231231 try to fix lost messages
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_DELETE_LOGFILE:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							int savelog;										/* saves LoggingToFile setting */
							savelog=LoggingToFile;								/* save it */
							if ( LoggingToFile ) CloseLogfileAndStopLogging();					/* close down */
							urc=DosDelete(NameOfLogFile);			/* zap it */
							if ( savelog )
								OpenLogFileAndTurnOnLogging();				/* restart logging */
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_LOG_NEW_LISTBOX_LINES:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							if ( LoggingToFile )
								CloseLogfileAndStopLogging();
							else
								OpenLogFileAndTurnOnLogging();
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_ADD_TIMESTAMP_TO_LOGGED_LINES:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							LogTimeWithMsgToFile = !LogTimeWithMsgToFile;
							PutProfileBool(KEY_ADD_TIME_TO_LOGGED_LINES, LogTimeWithMsgToFile);
							MenuWinCheckMenuItem(IDMI_ADD_TIMESTAMP_TO_LOGGED_LINES, LogTimeWithMsgToFile);
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_ADD_TIME_TO_LISTBOX_ENTRIES:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							/*--- Toggle state of flag -------------------------------*/
							AddTimeToDisplayedLines = ! AddTimeToDisplayedLines;

							/*--- Write to profile -----------------------------------*/
							PutProfileBool(KEY_TIME_IN_LISTBOX, AddTimeToDisplayedLines);

							/*--- Update Check state ---------------------------------*/
							MenuWinCheckMenuItem(IDMI_ADD_TIME_TO_LISTBOX_ENTRIES, AddTimeToDisplayedLines);
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_CASE_SENSITIVE_SEARCHES:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							/*--- Toggle state of flag -------------------------------*/
							CaseSensitiveSearches = ! CaseSensitiveSearches;

							/*--- Write to profile -----------------------------------*/
							PutProfileBool(KEY_CASE_SENSITIVE_SEARCHES, CaseSensitiveSearches);

							/*--- Update Check state ---------------------------------*/
							MenuWinCheckMenuItem(IDMI_CASE_SENSITIVE_SEARCHES, CaseSensitiveSearches);
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_RESTORE_PRESPARMS:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							/*--- Toggle state of flag -------------------------------*/
							RestorePresParms = ! RestorePresParms;

							/*--- Write to profile -----------------------------------*/
							PutProfileBool(KEY_RESTORE_PRESPARMS, RestorePresParms);

							/*--- Update Check state ---------------------------------*/
							MenuWinCheckMenuItem(IDMI_RESTORE_PRESPARMS, RestorePresParms);
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_SAVE_PRESPARMS:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							/*--- Toggle state of flag -------------------------------*/
							SavePresParms = ! SavePresParms;

							/*--- Write to profile -----------------------------------*/
							PutProfileBool(KEY_SAVE_PRESPARMS, SavePresParms);

							/*--- Update Check state ---------------------------------*/
							MenuWinCheckMenuItem(IDMI_SAVE_PRESPARMS, SavePresParms);
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_ADD_COMMENT:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							SHORT InsertCommentLocation;

							/*--- Work out where to insert the comment ---------------*/
							if ( NumberOfSelectedItemsInListBox(&InsertCommentLocation, NULL) > 1 )
								{
								/*--- Too many selected line (should be 0 or 1) ------*/
								MsgBox(ATTR_ERROR, "TOO MANY SELECTED LINES",
											 "You should indicate where to place the comment "
											 "by selecting the line after which you want the "
											 "comment placed."
											 "\n\n"
											 "If no lines are selected then the comment is "
											 "added to the end."
											);
								break;

								}

							/*--- Get a comment from the user ------------------------*/
							stuff[0] = '\0';
							if ( ! AskQuestion("Please enter the Comment you wish to add", stuff, 101, NULL, TRUE) )
								break;									//User cancelled  request

							/*--- Determine where the comment should be placed -------*/
							if ( InsertCommentLocation == LIT_NONE )
								InsertCommentLocation = LIT_END;				 //Insert at end
							else
								{
								/*--- We need to deselect the current line -----------*/
								WinSendMsg(hListBox, LM_SELECTITEM, MPFROMSHORT(InsertCommentLocation), MPFROMSHORT(FALSE));
								InsertCommentLocation++;								 //Insert after selected item
								}

							/*--- Add the comment (remember its position) ------------*/
							InsertCommentLocation = AddLineToListbox(stuff, 0, InsertCommentLocation);

							/*--- Select the new comment -----------------------------*/
							WinSendMsg(hListBox, LM_SELECTITEM, MPFROMSHORT(InsertCommentLocation), MPFROMSHORT(TRUE));
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_MAX_LINES_IN_LISTBOX:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							char    NumberString[6];
							ULONG   Value;
							char  * Ptr;

							/*--- Ask for the search string --------------------------*/
							sprintf(NumberString, "%d", MaxLinesKeptInWindow);
							if ( ! AskQuestion("Restrict listbox to how many lines?", NumberString, sizeof(NumberString), DEFAULT_NUMBER_OF_LISTBOX_ITEMS_STRING, TRUE) )
								break;									//User cancelled  request

							/*--- Get value and make sure its valid ------------------*/
							Value = strtoul(NumberString, &Ptr, 10);
							if ( (Value == 0) || (Value > 32700) || (*Ptr != '\0') )
								{
								/*--- Invalid number ---------------------------------*/
								MsgBox(ATTR_ERROR, "INVALID NUMBER OF LINES",
											 "We expected a non zero valid decimal value, the "
											 "number \"%s\" does not match this requirement!"
											 ,NumberString
											);
								}
							else
								{
								/*--- Number valid so update global variable etc -----*/
								MaxLinesKeptInWindow = Value;
								PutProfileUlong(KEY_MAX_LINES_IN_LISTBOX, MaxLinesKeptInWindow);
								UpdateMaxLinesInListBoxMenuText();
								}
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_FIND:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							char  OriginalSearchString[sizeof(SearchString)];

							/*--- Ask for the search string --------------------------*/
							strcpy(OriginalSearchString, SearchString);
							if ( ! AskQuestion("Search for?", SearchString, sizeof(SearchString), OriginalSearchString, FALSE) )
								break;									//User cancelled  request

							/*--- If changes then write it to the INI file -----------*/
							if ( strcmp(SearchString, OriginalSearchString) != 0 )
								WriteProfileString(KEY_SEARCH_STRING, SearchString);

							/*--- This is not a find first operation -----------------*/
							LastMatchLocation = LIT_NONE;
							FindNext();
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_HELP_KEYS:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							DisplaySpecificHelpPanel(IDHLP_PMPRINTF_EXE_KEYS);
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_HELP_CONTENTS:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							DisplayHelpContents();
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_FINDNEXT:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							FindNext();
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_FINDPREV:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							FindPrevious();
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_PAUSE_RESUME:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							TogglePausedState();
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_BEEP_ON_MSG_ARRIVAL:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							/*--- This option is not saved to profile ----------------*/
							BeepOnMessageArrival = ! BeepOnMessageArrival;
							MenuWinCheckMenuItem(IDMI_BEEP_ON_MSG_ARRIVAL, BeepOnMessageArrival);
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_EMPTY_LISTBOX:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							WinSendMsg(hListBox, LM_DELETEALL, NULL, NULL);
							LinesInListbox=0;
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_START_FONT_PALETTE:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							StartWpsObject("<WP_FNTPAL>", "FONT PALETTE");
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_START_SCOLOR_PALETTE:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							StartWpsObject("<WP_LORESCLRPAL>", "SOLID COLOR PALETTE");
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_START_MCOLOR_PALETTE:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							StartWpsObject("<WP_HIRESCLRPAL>", "MIXED COLOR PALETTE");
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_SPECIFY_TIME_FMT:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							/*--- Get the time format from the user ------------------*/
							if ( ! AskQuestion("Specify the time format (see strftime())", LogTimeFormat, sizeof(LogTimeFormat), DEFAULT_TIME_FMT, FALSE) )
								break;									//User cancelled  request

							/*--- Save the time format in the profile ----------------*/
							WriteProfileString(KEY_TIME_FORMAT, LogTimeFormat);
							break;
						}
					/*++++++++++++++++++++++++++++++++++++++*/
					case IDMI_EXIT_PGM:
					/*++++++++++++++++++++++++++++++++++++++*/
						{
							WinPostMsg(hwnd, WM_QUIT, NULL, NULL);
							break;
						}
					}
				return((MRESULT)FALSE);
			}
		/********************/
		case UM_GOTLINE:
		/********************/
			{
				/*--- Add the line into th listbox -------------------------------*/
				AddLineToListbox((char  *)mp1, (time_t)mp2, LIT_END);

				/*--- Free the messages memory -----------------------------------*/
				// 20231230 AB is done in above in AddLineToListbox, SMP issue otherwise!!!
				// DosFreeMem((char  *)mp1);
				return(MRESULT)FALSE;
			}
		/********************/
		case WM_PAINT:
		/********************/
			{
				HPS hps;

				/*--- Makes painting occur faster --------------------------------*/
				hps = WinBeginPaint(hwnd, 0, 0);
				WinEndPaint(hps);
				return(MRESULT)FALSE;
			}
		/**************************************/
		case WM_ERASEBACKGROUND:
		/**************************************/
			{
				/*--- Ensure small ribbon above listbox (if any) gets painted ----*/
				return((MRESULT)TRUE);
			}
		/********************/
		case UM_FILETRUNC:
		/********************/
			{
				sprintf(stuff,
								"One or more lines were truncated at length %i while writing the log file '%s'",
								MAX_LINE_LENGTH, NameOfLogFile);
				WinMessageBox(HWND_DESKTOP, hFrame, stuff,
											"Truncation warning...", /* title */ 0,
											(unsigned)MB_CANCEL | (unsigned)MB_ICONEXCLAMATION);
				return(MRESULT)FALSE;
			}
		/**************************************/
		case UM_ERROR_WRITING_TO_LOG_FILE:
		/**************************************/
			{
				CloseLogfileAndStopLogging();											 /* Close and de-check */
				sprintf(stuff,
								"Sorry, error %i (from DosWrite) while writing the log file '%s'",
								SHORT1FROMMP(mp1), NameOfLogFile);
				WinMessageBox(HWND_DESKTOP, hFrame, stuff,
											"Logging ended...",	/* title */ 0,
											(unsigned)MB_CANCEL | (unsigned)MB_ICONEXCLAMATION);
				return(MRESULT)FALSE;
			}
		/**************************************/
		case UM_QUEUE_CREATED:
		/**************************************/
			{
				WinShowWindow(hFrame, TRUE);							 /* make main visible */
				WinSetActiveWindow(HWND_DESKTOP, hFrame);	 /* .. and on top */
				WinInvalidateRect(hListBox, NULL, 0);			 /* force repaint for V2 */
				return(MRESULT)FALSE;
			}
		/**************************************/
		case UM_ERROR_CANT_CREATE_QUEUE:
		/**************************************/
			{
				WinAlarm(HWND_DESKTOP, WA_ERROR);
				sprintf(stuff,"Sorry, error %i in creating the input queue",
								SHORT1FROMMP(mp1));
				WinMessageBox(HWND_DESKTOP, hFrame, stuff,
											"Cannot start PMprintf...",	/* title */
											0,
											(unsigned)MB_CANCEL | (unsigned)MB_ICONEXCLAMATION);
				WinPostMsg(hwnd, WM_QUIT, NULL, NULL);
				return(MRESULT)FALSE;
			}
		/**************************************/
		case UM_ERROR_CANT_READ_QUEUE:
		/**************************************/
			{
				WinAlarm(HWND_DESKTOP, WA_ERROR);
				sprintf(stuff,
								"Sorry, error %i (from DosReadQueue) in processing the input queue",
								SHORT1FROMMP(mp1));
				WinMessageBox(HWND_DESKTOP, hFrame, stuff,
											"Ending PMprintf...",	/* title */
											0,
											(unsigned)MB_CANCEL | (unsigned)MB_ICONEXCLAMATION);
				WinPostMsg(hwnd, WM_QUIT, NULL, NULL);
				return(MRESULT)FALSE;
			}
		/**************************************/
		case UM_ERROR_QUEUE_DUPLICATED:
		/**************************************/
			{

				char uqname[9];
				WinAlarm(HWND_DESKTOP, WA_ERROR);
				strcpy(uqname,CmdLineOptions.ShortQueueName); strupr(uqname);	 /* name in upper case please */
				sprintf(stuff, "Sorry - PmPrintf queue '%s' already in use", uqname);
				WinMessageBox(HWND_DESKTOP, hFrame,
											stuff,
											"Cannot start PMprintf...",	/* title */ 0,
											(unsigned)MB_CANCEL | (unsigned)MB_ICONEXCLAMATION);
				WinPostMsg(hwnd, WM_QUIT, NULL, NULL);
				return(MRESULT)FALSE;
			}
		/**************************************/
		case WM_CLOSE:
		case WM_SAVEAPPLICATION:
		/**************************************/
			{
				SavePositionAndPresentationParameters(FALSE);
				break;
			}
		/**************************************/
		case WM_DESTROY:
		/**************************************/
			{
				if ( LoggingToFile )
					DosClose(LogfileHandle);		/* ensure closed */
				return(MRESULT)FALSE;
			}
		}

	/*--- Do default processing ----------------------------------------------*/
	return( WinDefWindowProc(hwnd, msg, mp1, mp2) );
	}



/*===========================================================================*/
static void _System SyntaxError(int ExitRc, char  * Format, ...)
/*===========================================================================*/
	{
	va_list  Args;

	/*--- Beep ---------------------------------------------------------------*/
	DosBeep(1000,80);

	/*--- Display the error message in a PM popup ----------------------------*/
	va_start(Args, Format);
	MsgBoxVa(ATTR_ERROR, "PmPrintf SYNTAX ERROR", Format, Args);
	va_end(Args);

	/*--- Exit to dos --------------------------------------------------------*/
	exit(ExitRc);
	}

#define INCL_LOADEXCEPTQ                // needed for exceptq.h
#include "exceptq.h"

/*===========================================================================*/
int main(int argc, char **argv)
/*===========================================================================*/
	{
	EXCEPTIONREGISTRATIONRECORD exRegRec;
	QMSG    qmsg;
	int     Index;
	char  * ArgumentPtr;
	char sClassName[4 * MAXNAMEL];
	char sWinTitle[4 * MAXNAMEL];
	int iLen;
	ULONG   flCreateFlags = FCF_TITLEBAR
													| FCF_SYSMENU
													| FCF_SIZEBORDER
													| FCF_MINMAX
													| FCF_SHELLPOSITION
													| FCF_NOBYTEALIGN
													| FCF_ACCELTABLE
													| FCF_MENU
													| FCF_ICON;

	LoadExceptq(&exRegRec, "I", "PMPrintf Viewer build: "__DATE__" "__TIME__);

	/*--- Initialize PM ------------------------------------------------------*/
	hab=WinInitialize(0);
	hmq=WinCreateMsgQueue(hab,0);

	/*--- Work through the command line --------------------------------------*/
	for ( Index=1; (ArgumentPtr=argv[Index]) != NULL; Index++ )
		ProcessParameterDieOnFailure(&CmdLineOptions, NULL, SyntaxError, ArgumentPtr);

	/*--- Get name of queue in upper case ------------------------------------*/
	sprintf(FullQueueName, QUEUE_NAME_PREFIX "%s", CmdLineOptions.ShortQueueName);
	strupr(FullQueueName);

	/*--- Create some semaphores ---------------------------------------------*/
	DosCreateEventSem(NULL, &hSemPaused,  0L, TRUE);	//Start in unpaused state
	DosCreateEventSem(NULL, &hSemGotLine, 0L, FALSE);

	
	sprintf (sClassName, "%s", PMPRINTF_CLASS_NAME);
	iLen = strlen(sClassName);
	if (iLen	> MAX_CLASS_NAME_LENGTH) sClassName[MAX_CLASS_NAME_LENGTH] = '\0'; 	// 20231230 cut title if too long
	/*--- Register the PMPRINTF class ----------------------------------------*/
	WinRegisterClass(   hab,
											//PMPRINTF_CLASS_NAME,
											sClassName,
											PmprintfWindowProcedure,
											CS_SIZEREDRAW,
											0
									);

	sprintf (sWinTitle, "%s", PMPRINTF_CLASS_NAME);
	iLen = strlen(sWinTitle);
	if (iLen	> MAX_CLASS_NAME_LENGTH) sWinTitle[MAX_CLASS_NAME_LENGTH] = '\0'; 	// 20231230 cut title if too long
	/*--- Create the window --------------------------------------------------*/
	hFrame = WinCreateStdWindow( HWND_DESKTOP,
															 WS_SAVEBITS,
															 &flCreateFlags,
															 //PMPRINTF_CLASS_NAME,
															 sClassName,
															 //" - Console",
															 sWinTitle,
															 0L,
															 0,
															 ID_FRAMERC,
															 &hClient
														 );

	// 20231231 try to fix lost messages with higher priority
	DosSetPriority (	PRTYS_PROCESSTREE,   	/* Change this process tree */
										PRTYC_TIMECRITICAL,	  /* Time critical class    */
										25L,                	/* Increase by 15         */
										0L);                	/* Assume current thread  *
	
	/*--- Start message loop -------------------------------------------------*/
	while ( WinGetMsg(hab, &qmsg, 0, 0, 0) )
		WinDispatchMsg(hab,&qmsg);

	/*--- Destroy the help ---------------------------------------------------*/
	DestroyHelpInstance();

	/*--- Destroy the window -------------------------------------------------*/
	WinDestroyWindow(hFrame);

	/*--- Terminate PM -------------------------------------------------------*/
	WinDestroyMsgQueue(hmq);
	WinTerminate(hab);
	return(__LINE__);								//Return code is meaningless
	}


