/*---------------------------------------------------------------------------+
|
|     MODULE NAME:   PMPFLOG.C
|
|
|         $Author:   Dennis_Bareis  $
|                    Andreas Buchinger  2009-2024
|       $Revision:   1.0  $
|           $Date:   28 Sep 1996 15:08:52  $
|        $Logfile:   C:/PROJECTS/OS2/PVCS.IT/PMPRINTF/PMPFLOG.C_V  $
|
|     DESCRIPTION:   Stand alone program which logs received messages.
|
+---------------------------------------------------------------------------*/

/*------ Make constant string definition more efficient & protect them! -----*/
#pragma  strings(readonly)

/*------ Include required Header files --------------------------------------*/
#define INCL_DOSPROCESS
#define INCL_NOPMAPI
#define INCL_LONGLONG  // 20090705 AB
#define INCL_DOSERRORS

#include <os2.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include "FeCommon.H"

#include <float.h>

/*--- General Global Variables ----------------------------------------------*/
static CMDLINEOPTIONS      CmdLineOptions         = DEFAULT_CMDLINEOPTIONS;
static LOGCMDLINEOPTIONS   LogCmdLineOptions      = DEFAULT_LOGCMDLINEOPTIONS;
static char                FullQueueName[CCHMAXPATH];
#define QUEUE_NAME_PREFIX  "\\QUEUES\\"



/*===========================================================================*/
static void _System SyntaxError(int ExitRc, char  * Format, ...)
/*===========================================================================*/
	{
	va_list  Args;

	/*--- Beep ---------------------------------------------------------------*/
	DosBeep(1000,80);

	/*--- Display the correct syntax -----------------------------------------*/
	fprintf(stderr, "CORRECT SYNTAX\n"
					"~~~~~~~~~~~~~~\n"
					"\tPMPFLOG[.EXE]  /Switch[:Parameter]\n"
					"\n"
					"AVAILABLE SWITCHES\n"
					"~~~~~~~~~~~~~~~~~~\n"
					"\t* NewQueueName\n"
					"\t* LogFileName\n"
					"\t* LogTimeFmt    : Codes as documented for 'C' strftime()\n"
					"\n"
				 );


	/*--- Display the error message ------------------------------------------*/
	fprintf(stderr, "SYNTAX ERROR\n"
					"~~~~~~~~~~~~\n"
				 );
	va_start(Args, Format);
	vfprintf(stderr, Format, Args);
	va_end(Args);
	fprintf(stderr, ".\n");
	DosSleep(1500L);

	/*--- Exit to dos --------------------------------------------------------*/
	exit(ExitRc);
	}


/*===========================================================================*/
static void _System FatalError(int ExitRc, char  * Format, ...)
/*===========================================================================*/
	{
	va_list  Args;

	/*--- Beep ---------------------------------------------------------------*/
	DosBeep(1000,80);

	/*--- Display the error message ------------------------------------------*/
	fprintf(stderr, "FATAL ERROR\n"
					"~~~~~~~~~~~\n"
				 );
	va_start(Args, Format);
	vfprintf(stderr, Format, Args);
	va_end(Args);
	fprintf(stderr, ".\n");
	DosSleep(1500L);

	/*--- Exit to dos --------------------------------------------------------*/
	exit(ExitRc);
	}



/*===========================================================================*/
void _Optlink LogMsgFromQueue(char * Msg, time_t MsgTimeStamp, BOOL Beep, ULONG NumberMsgsInQueue)
/*===========================================================================*/
	{
	static HFILE   LogHandle = NULLHANDLE;
	static ULONG   MsgNumber = 0;
	ULONG   DosRc;
	char    FormattedTime[100];
	ULONG   CharsToWrite;
	ULONG   CharsWritten;
#define CRLF  "\r\n"
	APIRET  rc = NO_ERROR;
	ULONG ulMemSize = 1;
	ULONG ulMemFlags;
	char *HighMemBuffer;
	char cBuf[4];
	int i;
	char buf[1024];
	char szMessage[] = "ERROR: got invalid string, invalid memory ERROR at";

	/*--- Do we need to open the file? ---------------------------------------*/
	if ( LogHandle == NULLHANDLE )
		{
		ULONG  OpenAction;
		ULONG  FilePointer;

		/*--- We need to open the log file -----------------------------------*/
		DosRc = DosOpen(   LogCmdLineOptions.FullLogFileName,
											 &LogHandle,
											 &OpenAction,
											 0L,
											 FILE_NORMAL,
											 FILE_OPEN | FILE_CREATE,
											 OPEN_ACCESS_READWRITE
											 | OPEN_SHARE_DENYWRITE
											 | OPEN_FLAGS_SEQUENTIAL,
											 0L
									 );
		if ( DosRc != 0 )
			FatalError(__LINE__, "Can't open the log file \"%s\" (DosRc=%lu)", LogCmdLineOptions.FullLogFileName, DosRc);

		/*--- Now move the pointer to the end (so we append to log) ----------*/
		if ( (DosRc=DosSetFilePtr(LogHandle, 0L, FILE_END, &FilePointer)) != 0 )
			FatalError(__LINE__, "Can't move to EOF on the log file \"%s\" (DosRc=%lu)", LogCmdLineOptions.FullLogFileName, DosRc);
		}

	/*--- Get the time from the message & format it ------------------*/
	strftime(FormattedTime, sizeof(FormattedTime), LogCmdLineOptions.LogTimeFormat, localtime(&MsgTimeStamp));

	/*--- Write the time to the file ---------------------------------*/
	CharsToWrite = strlen(FormattedTime);
	DosRc = DosWrite(LogHandle, FormattedTime, CharsToWrite, &CharsWritten);
	if ( (DosRc != 0) || (CharsWritten != CharsToWrite) )
		FatalError(__LINE__, "Write failure on the log file \"%s\" (DosRc=%lu)", LogCmdLineOptions.FullLogFileName, DosRc);

	
	// now check for valid message and get it from HighMem
	rc = DosQueryMem(Msg, &ulMemSize, &ulMemFlags);		// 20181103 AB check for valid string
	if ( rc != NO_ERROR || !(ulMemFlags & PAG_READ) )
		{	// invalid mem
		sprintf(buf, "%s 0x%lX (%s, %s, %d)", szMessage,  Msg, __FILE__, __FUNCTION__, __LINE__);
		HighMemBuffer = buf;
		}
	else
		{	 // only check pointer to highmem when pointer in lower shared mem is valid
		memcpy(&HighMemBuffer, Msg, sizeof(char *) );
		rc = DosQueryMem(HighMemBuffer, &ulMemSize, &ulMemFlags);		// 20230702 AB check for valid mem pointer
		if ( rc != NO_ERROR || !(ulMemFlags & PAG_READ) )
			{	// invalid mem
			sprintf(buf, "%s 0x%lX (%s, %s, %d)", "PmPrintf ERROR: invalid HighMem buffer",  HighMemBuffer, __FILE__, __FUNCTION__, __LINE__);
			HighMemBuffer = buf;
			// try to concenate as it would be old style, none HighMem message
			//vsprintf(buf + strlen(buf), Msg, args);
			strcat(buf + strlen(buf), Msg);
			}
		}
	// clear the lower shared mem buffer (pointer to message in HighMem)
	DosFreeMem(Msg);
	
	/*--- Now write the information ------------------------------------------*/
	MsgNumber++;
	CharsToWrite = strlen(HighMemBuffer);
	DosRc = DosWrite(LogHandle, HighMemBuffer, CharsToWrite, &CharsWritten);
	if ( (DosRc != 0) || (CharsWritten != CharsToWrite) )
		FatalError(__LINE__, "Write failure on the log file \"%s\" (DosRc=%lu)", LogCmdLineOptions.FullLogFileName, DosRc);

	/*--- Output CR & LF -----------------------------------------------------*/
	CharsToWrite = sizeof(CRLF)-1;
	DosRc = DosWrite(LogHandle, CRLF, CharsToWrite, &CharsWritten);
	if ( (DosRc != 0) || (CharsWritten != CharsToWrite) )
		FatalError(__LINE__, "Write failure on the log file \"%s\" (DosRc=%lu)", LogCmdLineOptions.FullLogFileName, DosRc);

	/*--- Flush I/O if no more messages in the queue -------------------------*/
	if ( NumberMsgsInQueue == 0 )
		{
		/*--- Flush the I/O --------------------------------------------------*/
		DosResetBuffer(LogHandle);

		/*--- Indicate somethings happening ----------------------------------*/
		printf("%s#%lu\n", FormattedTime, MsgNumber);
		}

	/*--- Free the memory that the message was passed in! --------------------*/
	DosFreeMem(HighMemBuffer);
	}





/*===========================================================================*/
int main(int argc, char **argv)
/*===========================================================================*/
	{
	int     Index;
	char  * ArgumentPtr;
	time_t  StartingTime;
	char    FormattedTime[100];

	/*--- Display Startup Message --------------------------------------------*/
	printf(
				"[]------------------------------------------------------------[]\n"
				"| PMPFLOG v%s              |\n"
				"| (C)opyright Dennis Bareis 1996, Andreas Buchinger 2009-2024  |\n"
				"| All Rights Reserved.                                         |\n"
				"[]------------------------------------------------------------[]\n"
				"\n"
				, PGM_VERSION
				);

	/*--- Work through the command line --------------------------------------*/
	for ( Index=1; (ArgumentPtr=argv[Index]) != NULL; Index++ )
		ProcessParameterDieOnFailure(&CmdLineOptions, &LogCmdLineOptions, SyntaxError, ArgumentPtr);

	/*--- Get name of queue in upper case ------------------------------------*/
	sprintf(FullQueueName, QUEUE_NAME_PREFIX "%s", CmdLineOptions.ShortQueueName);
	strupr(FullQueueName);

	/*--- Output current Information -----------------------------------------*/
	time(&StartingTime);
	strftime(FormattedTime, sizeof(FormattedTime), LogCmdLineOptions.LogTimeFormat, localtime(&StartingTime));
	printf(
				"Queue = %s\n"
				"Log   = %s\n"
				"Time  = \"%s\"\n"
				"\n"
				,FullQueueName
				,LogCmdLineOptions.FullLogFileName
				,FormattedTime
				);

	/*--- Create the Queue and log any messages received ---------------------*/
	CreatePmprintfQueue(FullQueueName, LogMsgFromQueue, FatalError);

	/*--- Thats all Folks ----------------------------------------------------*/
	return(__LINE__);								//Return code is meaningless
	}

