/*---------------------------------------------------------------------------+
|
|     (C)opyright Dennis Bareis and Andreas Buchinger
|     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
|
|     MODULE NAME:   PMPRINTF.C
|
|
|         $Author:   Dennis_Bareis  $
|                    Andreas Buchinger  2009-2024
|       $Revision:   1.2  $
|           $Date:   28 Sep 1996 15:08:26  $
|        $Logfile:   C:/PROJECTS/OS2/PVCS.IT/PMPRINTF/PMPRINTF.C_V  $
|
|     DESCRIPTION:   This module contains routines to transmit data to
|                    PMPRINT.EXE.  Very little stack space is used by these
|                    functions.
+---------------------------------------------------------------------------*/

/*------ Make constant string definition more efficient & protect them! -----*/
#pragma  strings(readonly)

/*------ Include required Header files --------------------------------------*/
#define  INCL_DOSPROCESS
#define  INCL_DOSQUEUES
#define  INCL_DOSERRORS
#define  INCL_NOPMAPI
#define  INCL_LONGLONG  // 20090705 AB

#include <os2.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <stdarg.h>
#include "PMPRINTF.H"
#include "FECOMMON.H"		// needed for PGM_VERSION


/*--- Allocate the pointer which will point to thread data ------------------*/
static PER_THREAD   **ThreadInfo = NULL;

/*--- Set the initial mode --------------------------------------------------*/
static PMPRINTF_MODE PmprintfMode = PMPRINTF_DEFAULT_MODE;

/*--- Set up the name of the queue this process uses ------------------------*/
static char * QueueNameThisProcess = PMPRINTF_DEFAULT_QUEUE_NAME;

/*--- DB$ First Failure Technology! -----------------------------------------*/
static ULONG   HadError       = 0;				 //Boolean, != 0 means error (value is count of errors)
static BOOL    BeingReported  = FALSE;
static ULONG   ErrorLine      = 0;
static char  * ErrorDoingWhat = NULL;
static ULONG   ErrorType      = 0;
static ULONG   ErrorRc        = 0;
static time_t  ErrorTime      = 0;
#define FMT_DATETIME "%a"    /* Day of Week (Fri)  */  \
                                    " "                               \
                                    "%H"    /* Time-HH            */  \
                                    ":"                               \
                                    "%M"    /* Time-MM            */  \
                                    ":"                               \
                                    "%S"    /* Time-SS            */  \
                                    "%p"    /* Time AM or PM      */
static PIB   * ErrorPib       = NULL;
static TIB   * ErrorTib       = NULL;





/*--- Throw away macros -----------------------------------------------------*/
#define _PMPRINTF_MAKE_STR(NonString)   _PMPRINTF_MAKE_STR2(NonString)
#define _PMPRINTF_MAKE_STR2(NonString)  #NonString
#define _BitsAreSet(   field,mask)      ( (field) & (mask) )
#define _NoBitsAreSet( field,mask)      ( ( (field) & (mask) ) == 0 )




/*===========================================================================*/
static void Noise(void)
/*===========================================================================*/
	{
	DosBeep(1000,80);
	DosBeep(1500,80);
	DosBeep(2000,80);
	DosBeep(2500,80);
	DosBeep(2000,80);
	DosBeep(1500,80);
	DosBeep(1000,80);
	}



/*===========================================================================*/
static ULONG _Optlink HadAnError(   ULONG        LineNumber,
																		char       * DoingWhat,
																		ULONG        RcType,
#define RC_OS2   '2'
#define RC_C     'C'
#define RC_OTHER  0
																		ULONG        Rc,
																		PER_THREAD * PerThread
																)
/*===========================================================================*/
	{
	/*--- Keep count of errors (only record information & beep on 1st) ------*/
	if ( HadError++ == 0 )
		{
		/*--- This is the first error, so record details --------------------*/
		BeingReported  = FALSE;
		ErrorLine      = LineNumber;
		ErrorDoingWhat = DoingWhat;
		ErrorType      = RcType;
		ErrorRc        = Rc;
		time(&ErrorTime);
		if ( PerThread == NULL )
			{
			ErrorPib   = NULL;
			ErrorTib   = NULL;
			}
		else
			{
			ErrorPib   = PerThread->Pib;
			ErrorTib   = PerThread->Tib;
			}

		/*--- Make a fairly distinctive noise! ------------------------------*/
		if ( _BitsAreSet(PmprintfMode, PMPRINTF_NOISE_ON_1ST_ERROR) )
			Noise();
		}
	return(LineNumber);
	}


/*===========================================================================*/
static PER_THREAD * _Optlink GetThreadMemory(void)
/*===========================================================================*/
	{
	PER_THREAD * ThreadData;
	PIB        * Pib;
	TIB        * Tib;
	TID          ThisThreadId;
	ULONG        DosRc;

	/*--- Get thread information --------------------------------------------*/
	if ( (DosRc=DosGetInfoBlocks(&Tib, &Pib)) != 0 )
		{
		HadAnError(__LINE__, "DosGetInfoBlocks()", RC_OS2, DosRc, NULL);
		return(NULL);
		}
	ThisThreadId  = Tib->tib_ptib2->tib2_ultid;

	/*--- Make sure that the number of the thread is not too high -----------*/
	if ( ThisThreadId > MAX_THREADS )				 //ThreadId less 1 is used as index!
		{
		HadAnError(__LINE__, "\"More than " _PMPRINTF_MAKE_STR(MAX_THREADS) " threads\"", RC_OTHER, (ULONG)ThisThreadId, NULL);
		return(NULL);
		}

	/*--- Allocate the thread info buffer if we need to ---------------------*/
	if ( ThreadInfo == NULL )
		{
		ULONG   DataLength = MAX_THREADS * sizeof(PER_THREAD *);

		/*--- Allocate & clear the thread data pointers ---------------------*/
		if ( (ThreadInfo=malloc(DataLength)) == NULL )
			{
			/*--- Could not allocate the memory, inform user and bail out ---*/
			HadAnError(__LINE__, "malloc()", RC_C, errno, NULL);
			return(NULL);
			}
		memset(ThreadInfo, 0, DataLength);				//Set all pointers to NULL
		}

	/*--- Point to structure (allocate if required) -------------------------*/
	if ( (ThreadData=ThreadInfo[ThisThreadId-1]) == NULL )
		{
		/*--- Allocate a Thread info structure ------------------------------*/
		if ( (ThreadData=malloc(sizeof(PER_THREAD))) == NULL )
			{
			/*--- Could not allocate the memory, inform user and bail out ---*/
			HadAnError(__LINE__, "malloc()", RC_C, errno, NULL);
			return(NULL);
			}
		ThreadInfo[ThisThreadId-1] = ThreadData;	//Update thread pointer array

		/*--- Initialize the newly allocated threads memory -----------------*/
		ThreadData->MacroHeaderLength = 0;
		ThreadData->PmprintfDisabled  = (NULL != getenv(ENVVAR_NOPMPRINTF_BY_DEFAULT_IF_EXISTS));
		}

	/*--- Set up the Thread & Process information ---------------------------*/
	ThreadData->Pib = Pib;
	ThreadData->Tib = Tib;

	/*--- Return the pointer to the caller ----------------------------------*/
	return(ThreadData);
	}



/*===========================================================================*/
static ULONG _Optlink SendLineToPmprintf(PER_THREAD *PerThread, long DataLength)
/*                                                                           */
/*  This routine is not called directly by the user but is used internally   */
/*  to output a line of data.                                                */
/*===========================================================================*/
	{
	static HQUEUE   QueueHandle = NULLHANDLE;
	static ULONG    ServersPID  = 0;
	int      TryCount;
	ULONG    DosRc;
	time_t   CurrentTime;
	void   * AllocatedMemory;
	void   * AllocatedMemory2;
	char   * LineHeaderPtr;
	ULONG    LineHeaderLength;

	/*--- Get the current time ----------------------------------------------*/
	time(&CurrentTime);

	/*--- Set up any "standard" header we require ---------------------------*/
	if ( PerThread->Line[0] == '\0' )
		{
		/*--- We have a blank line so we won't add a header -----------------*/
		PerThread->LineHeader[0] = '\0';
		LineHeaderLength         = 0;
		}
	else
		{
		/*--- Output the date/time? -----------------------------------------*/
		LineHeaderPtr = PerThread->LineHeader;
		if ( _BitsAreSet(PmprintfMode, PMPRINTF_SEND_TIME) )
			{
			/*--- We wish to output the time --------------------------------*/
			strftime(LineHeaderPtr, 30, FMT_DATETIME, localtime(&CurrentTime));
			LineHeaderPtr   += strlen(LineHeaderPtr);
			*LineHeaderPtr++ = ' ';			//Output a space
			}

		/*--- Output Process info? ------------------------------------------*/
		if ( _BitsAreSet(PmprintfMode, PMPRINTF_SEND_PROC_INFO) )
			{
			/*--- We wish to output the time --------------------------------*/
			LineHeaderPtr += sprintf(LineHeaderPtr, "P%03d,T%d"
															 ,((PIB *)PerThread->Pib)->pib_ulpid
															 ,((TIB *)PerThread->Tib)->tib_ptib2->tib2_ultid
															);
			*LineHeaderPtr++ = ' ';			//Output a space
			}

		/*--- Terminate the header with a colon if there is one -------------*/
		*LineHeaderPtr = '\0';
		if ( PerThread->LineHeader[0] != '\0' )
			strcat(LineHeaderPtr, ": ");
		LineHeaderLength = strlen(PerThread->LineHeader);
		}
	DataLength += LineHeaderLength;

	/*--- Create a retry loop -----------------------------------------------*/
	TryCount = 0;
	for ( ;; )
		{
		/*--- Open the queue (if not already open) --------------------------*/
		if ( ServersPID == 0 )
			{
			/*--- Open the queue --------------------------------------------*/
			if ( (DosRc=DosOpenQueue(&ServersPID, &QueueHandle, QueueNameThisProcess)) != 0 )
				return(HadAnError(__LINE__, "DosOpenQueue()", RC_OS2, DosRc, PerThread));
			}

		/*--- Allocate some memory we can GIVE to the server (place data in it) (20230702 AB) in highmem */
		if ( (DosRc=DosAllocSharedMem(&AllocatedMemory2, NULL, DataLength, OBJ_GIVEABLE|PAG_WRITE|PAG_COMMIT | OBJ_ANY)) != 0 )
			return(HadAnError(__LINE__, "DosAllocSharedMem()", RC_OS2, DosRc, PerThread));
		// allocate memory in lower shared mem for the pointer to the real message in highmem
		if ( (DosRc=DosAllocSharedMem(&AllocatedMemory, NULL, sizeof(void *), OBJ_GIVEABLE|PAG_WRITE|PAG_COMMIT )) != 0 )
			return(HadAnError(__LINE__, "DosAllocSharedMem()_2", RC_OS2, DosRc, PerThread));


		/*--- Give The memory to the server ---------------------------------*/
		if ( (DosRc=DosGiveSharedMem(AllocatedMemory, ServersPID, PAG_READ)) == 0 &&
				 (DosRc|=DosGiveSharedMem(AllocatedMemory2, ServersPID, PAG_READ)) == 0 )
			break;
		else
			{
			/*--- If we die here the server may have died, it may have restarted ---*/
			DosFreeMem(AllocatedMemory);
			DosFreeMem(AllocatedMemory2);
			if ( ++TryCount > 1 )
				return(HadAnError(__LINE__, "DosGiveSharedMem()", RC_OS2, DosRc, PerThread));
			else
				{
				/*--- Reset the server PID and try to open the queue --------*/
				ServersPID = 0;
				continue;
				}
			}
		}

	/*--- Send the data to PMPRINTF.EXE -------------------------------------*/
	if ( PerThread->Beep )
		{
		/*--- Reset beep state and tell PMPRINTF.EXE to beep when it gets msg ---*/
		PerThread->Beep = FALSE;
		DataLength      = - DataLength;
		}
	strcpy(AllocatedMemory2, PerThread->LineHeader);
	strcat(AllocatedMemory2, PerThread->Line);
	memcpy(AllocatedMemory, &AllocatedMemory2, sizeof(void *) );
	if ( (DosRc=DosWriteQueue(QueueHandle, CurrentTime, 4, AllocatedMemory, 0)) != 0 )	 // <---- adapt lengtht
		{
		/*--- We failed make noise! -----------------------------------------*/
		DosFreeMem(AllocatedMemory);
		DosFreeMem(AllocatedMemory2);
		return(HadAnError(__LINE__, "DosWriteQueue()", RC_OS2, DosRc, PerThread));
		}

	/*--- Free The memory ---------------------------------------------------*/
	if ( (DosRc=DosFreeMem(AllocatedMemory)) != 0 )
		return(HadAnError(__LINE__, "DosFreeMem()", RC_OS2, DosRc, PerThread));
// ?	Do not free the HighMem buffer here cause it will be needed until written by the viewer
	if ( (DosRc=DosFreeMem(AllocatedMemory2)) != 0 )
		return(HadAnError(__LINE__, "DosFreeMem()", RC_OS2, DosRc, PerThread));

	/*--- We successfully sent the data, if there is an error report it! ----*/
	if ( HadError )
		{
		/*--- Has another thread beat us to it? -----------------------------*/
		if ( ! BeingReported )
			{
			PID     Pid;
			TID     Tid;
			char  * ProcessCmdLine;
			char  * ErrorTypePtr;
			char    TimeOfError[20];

			/*--- Immediately set being reported state! ---------------------*/
			BeingReported = TRUE;

			/*--- Get Process information (if known) ------------------------*/
			if ( ErrorPib  == NULL )
				{
				Pid            = 0;
				Tid            = 0;
				ProcessCmdLine = "?";
				}
			else
				{
				Pid            = ErrorPib->pib_ulpid;
				Tid            = ErrorTib->tib_ptib2->tib2_ultid;
				ProcessCmdLine = ErrorPib->pib_pchcmd;
				}

			/*--- What sort of error was it ---------------------------------*/
			if ( ErrorType == 0 )
				ErrorTypePtr = "other return code";
			else
				{
				if ( ErrorType == RC_C )
					ErrorTypePtr = "'C' errno";
				else
					ErrorTypePtr = "OS/2 return code";
				}

			/*--- Format the current date/time (and then point to null byte) ----*/
			strftime(TimeOfError, sizeof(TimeOfError), FMT_DATETIME, localtime(&ErrorTime));

			/*--- Now set up structure and call ourselves to send -----------*/
			PerThread->Beep = TRUE;
			sprintf(PerThread->Line, "!!!! DATA LOST (%lu errors since %s) - %.100s failed at "
							"line #%d with "
							"%s of "
							"%lu decimal "
							"(PID=%lu, TID=%lu, Process=%.250s) "
							"!!!!"
							,HadError
							,TimeOfError
							,ErrorDoingWhat
							,ErrorLine
							,ErrorTypePtr
							,ErrorRc
							,Pid, Tid, ProcessCmdLine
						 );

			/*--- Now call ourselves to send data to PMPRINTF.EXE -----------*/
			if ( SendLineToPmprintf(PerThread, strlen(PerThread->Line)+1) == 0 )
				{
				/*--- Reset flag last to retain error even if reporting fails! --*/
				HadError = FALSE;


				/*--- Make a fairly distinctive noise! ------------------------------*/
				if ( _BitsAreSet(PmprintfMode, PMPRINTF_NOISE_ON_OK_AGAIN) )
					Noise();
				}
			}
		}

	/*--- Worked! -----------------------------------------------------------*/
	return(0);
	}



/*===========================================================================*/
static ULONG _Optlink SendCallersLineToPmprintf(PER_THREAD *PerThread)
/*                                                                           */
/*  The string to be printed is in the "FormattedMessage" variable, newlines */
/*  and beeps etc need processing.                                           */
/*===========================================================================*/
	{
	char   * SPtr                       = PerThread->FormattedMessage;
	char   * DPtr                       = PerThread->Line;
	BOOL     AlreadySentBeepForThisLine = FALSE;
	char     Character;
	long     CharacterCount;
	ULONG    Rc;
	BOOL     SendLine;

	/*--- Initialize parts of per thread structure --------------------------*/
	PerThread->Beep = FALSE;

	/*--- Process the data --------------------------------------------------*/
	CharacterCount = 0;
	SendLine       = FALSE;
	do
		{
		/*--- Get a character -----------------------------------------------*/
		Character = *SPtr++;

		/*--- Is this a special character -----------------------------------*/
		switch ( Character )
			{
			/*+++++++++++++++*/
			case '\a':
				/*+++++++++++++++*/
				/*--- Cause PMPRINTF.EXE to beep when it gets this message --*/
				if ( ! AlreadySentBeepForThisLine )
					{
					PerThread->Beep            = TRUE;
					AlreadySentBeepForThisLine = TRUE;
					}
				break;
				/*+++++++++++++++*/
			case '\r':
				/*+++++++++++++++*/
				/*--- Ignore carriage returns -------------------------------*/
				break;
				/*+++++++++++++++*/
			case '\n':
			case '\0':
				/*+++++++++++++++*/
				/*--- We have found the end of a line so output it ----------*/
				SendLine = TRUE;
				break;
				/*+++++++++++++++*/
			case '\t':
				/*+++++++++++++++*/
				/*--- Convert character and flow into "default" processing --*/
				Character = ' ';
				/*+++++++++++++++*/
			default:
				/*+++++++++++++++*/
				/*--- Save the character in the buffer ----------------------*/
				*DPtr++ = Character;
				CharacterCount++;
			}

		/*--- End of Line or Buffer full? -----------------------------------*/
		if ( SendLine || (CharacterCount >= PMPRINTF_MAX_LINE_CHARS) )
			{
			/*--- Flush the line --------------------------------------------*/
			*DPtr = '\0';
			if ( (Rc=SendLineToPmprintf(PerThread, CharacterCount+1)) != 0 )
				return(Rc);									//Error, its already beeped!

			/*--- Reset state for next line ---------------------------------*/
			SendLine = FALSE;
			CharacterCount = 0;
			DPtr           = PerThread->Line;
			}
		}
	while ( Character != '\0' );

	/*--- No errors occurred ------------------------------------------------*/
	return(0);
	}


/*===========================================================================*/
PMPRINTF_MODE _System  PmPrintfModeSet(PMPRINTF_MODE NewMode)
/*===========================================================================*/
	{
	PMPRINTF_MODE OriginalMode = PmprintfMode;

	/*--- Set the new mode --------------------------------------------------*/
	PmprintfMode = NewMode;

	/*--- return the original mode ------------------------------------------*/
	return(OriginalMode);
	}



/*===========================================================================*/
char * _System PmPrintfQueueNameThisProcess(char * StaticPmprintfQueueName)
/*                                                                           */
/* The caller should supply a queue name in "static" memory (as its not      */
/* copied!) and will receive the previous queue name as a return code.       */
/*                                                                           */
/* The queue name is used for all future queue operation for the whole       */
/* process.                                                                  */
/*                                                                           */
/* Note that you must call this function before the 1st write to the queue,  */
/* otherwise you will need to terminate the PMPRINTF.EXE program which       */
/* scans the original queue, it should "recover" to the new queue!           */
/*===========================================================================*/
	{
	char * OriginalQueueName = QueueNameThisProcess;
	char sBuffer[256];

	// 20090307 AB, dummy declaration to include floating point library
	float fTest = 1.3423;
	sprintf(sBuffer, "float test %f, longlong test %Ld\n", fTest * fTest / 1.00232);
	/*--- Set the new queue name --------------------------------------------*/
	if ( StaticPmprintfQueueName != NULL )
		{
		/*--- We are being asked to set a new queue name (is it valid?) -----*/
		if ( strncmp(StaticPmprintfQueueName, PMPRINTF_QUEUE_PREFIX, sizeof(PMPRINTF_QUEUE_PREFIX)-1) != 0 )
			return(NULL);									//Invalid Q name passed!
		QueueNameThisProcess = StaticPmprintfQueueName;
		}

	/*--- Return the previous queue name to the caller ----------------------*/
	return(OriginalQueueName);
	}



/*===========================================================================*/
ULONG _System I_PmpfSetMacroHeader(char  *SourceModule, int LineNumber, char  *Function)
/*                                                                           */
/* This is called via macros to cause the generation of a header in front of */
/* a message, the "MacroHeaderLength" "flag" is set to indicate its presence.     */
/*===========================================================================*/
	{
	PER_THREAD * PerThread;
	char       * ModuleName;

	/*--- Get allocated memory for this particular thread -------------------*/
	if ( (PerThread=GetThreadMemory()) == NULL )
		return(__LINE__);												//Error, its already beeped!
	if ( PerThread->PmprintfDisabled )
		return(0);															//Pmprintf disabled

	/*--- See if the source name has a path, remove if so -------------------*/
	ModuleName = SourceModule;
	if ( (ModuleName=strrchr(ModuleName, '\\')) != NULL )
		ModuleName++;			 //Point past '\'
	else
		ModuleName = SourceModule;

	/*--- Create the header -------------------------------------------------*/
	if ( Function == NULL )
		PerThread->MacroHeaderLength = sprintf(PerThread->FormattedMessage, "%s,%4d : ",          ModuleName, LineNumber);
	else
		PerThread->MacroHeaderLength = sprintf(PerThread->FormattedMessage, "%s,%4d, %.50s() : ", ModuleName, LineNumber, Function);

	return(0);
	}




/*===========================================================================*/
ULONG _System PmPrintfDisplayInterfaceVersionInfo(void)
/*                                                                           */
/* When called this routine will output the version information of this      */
/* interface.                                                                */
/*===========================================================================*/
	{
	PER_THREAD * PerThread;

	/*--- Get allocated memory for this particular thread -------------------*/
	if ( (PerThread=GetThreadMemory()) == NULL )
		return(__LINE__);												//Error, its already beeped!
	if ( PerThread->PmprintfDisabled )
		return(0);															//Pmprintf disabled

	/*--- Set up the message to be displayed --------------------------------*/
	strcpy(PerThread->Line, "**** PMPRINTF.C version " PGM_VERSION
//				 ", Last modified at "          __TIMESTAMP__
				 ", compiled at "               __DATE__ " " __TIME__
				 "  ****"
				);

	/*--- Now send this formatted line to PMPRINTF.EXE ----------------------*/
	return( SendLineToPmprintf(PerThread, strlen(PerThread->Line)+1) );
	}



/*===========================================================================*/
ULONG _System PmPrintfVa(char *Format, va_list args)
/*===========================================================================*/
	{
	PER_THREAD * ThisThreadsData;
	ULONG        Rc;
	
	/*--- Get allocated memory for this particular thread -------------------*/
	if ( (ThisThreadsData=GetThreadMemory()) == NULL )
		return(__LINE__);												//Error, its already beeped!
	if ( ThisThreadsData->PmprintfDisabled )
		return(0);															//Pmprintf disabled

	/*--- Format the passed message -----------------------------------------*/
	{
		vsprintf(ThisThreadsData->FormattedMessage+ThisThreadsData->MacroHeaderLength, Format, args);
		ThisThreadsData->MacroHeaderLength = 0;
	}
	/*--- Call the send routine & exit --------------------------------------*/
	Rc = SendCallersLineToPmprintf(ThisThreadsData);
	return(Rc);
	}



/*===========================================================================*/
ULONG _System PmPrintf(char *Format, ...)
/*===========================================================================*/
	{
	va_list  args;
	ULONG    Rc;

	/*--- Get pointer to variable arguments ---------------------------------*/
	va_start(args, Format);

	/*--- Output message ----------------------------------------------------*/
	Rc = PmPrintfVa(Format, args);

	/*--- Clean up and exit -------------------------------------------------*/
	va_end(args);
	return(Rc);
	}




/*===========================================================================*/
ULONG _System PmPrintfString(char *String)
/*===========================================================================*/
	{
	PER_THREAD * ThisThreadsData;
	ULONG        Rc;

	/*--- Get allocated memory for this particular thread -------------------*/
	if ( (ThisThreadsData=GetThreadMemory()) == NULL )
		return(__LINE__);												//Error, its already beeped!
	if ( ThisThreadsData->PmprintfDisabled )
		return(0);															//Pmprintf disabled

	/*--- Simply copy the passed data to the buffer -------------------------*/
	strcpy(ThisThreadsData->FormattedMessage+ThisThreadsData->MacroHeaderLength, String);
	ThisThreadsData->MacroHeaderLength = 0;

	/*--- Call the send routine & exit --------------------------------------*/
	Rc = SendCallersLineToPmprintf(ThisThreadsData);
	return(Rc);
	}


/******************************************************************************
ULONG _System PmPrintfGetVersion(int ReservedForFutureUse);
	Returns current version of PmPrintf package
	Useful to decide if HighMem string transfer mode is supported (since 2.70)
	Version string coded as vMajor.vMinor.vBeta (deliminator .). Only 2 digits
	per field are allowed.
	
	Returns version coded as ULONG with 2 digits per field f.i. 027100 means
	version 2.71.00 (vMajor.vMinor.vBeta)
	
	int ReservedForFutureUse - is currently no used
******************************************************************************/
ULONG _System PmPrintfGetVersion(int ReservedForFutureUse)
	{
	#define MAX_VERSIONSTRING_CHARS 	16
	
	char buf[MAX_VERSIONSTRING_CHARS];
	char buf2[MAX_VERSIONSTRING_CHARS];
	char *pP = buf;
	char *pP2 = buf2;
	int iVerMajor;
	int iVerMinor;
	int iVerBeta;	
	int i = 0;
	
	strncpy (buf, PGM_VERSION, MAX_VERSIONSTRING_CHARS);
	//pP = buf;
	
	while (*pP != '.' && *pP != ' ' && i < MAX_VERSIONSTRING_CHARS)
		{
		*pP2++ = *pP++;
		i++;
		}
	*pP2 = '\0';
	pP++;
	iVerMajor = atoi(buf2);
	pP2 = buf2;	
	
	while (*pP != '.' && *pP != ' ' && i < MAX_VERSIONSTRING_CHARS)
		{
		*pP2++ = *pP++;
		i++;
		}
	*pP2 = '\0';
	pP++;
	iVerMinor = atoi(buf2);
	pP2 = buf2;	
	
	while (*pP != '.' && *pP != ' ' && i < MAX_VERSIONSTRING_CHARS)
		{
		*pP2++ = *pP++;
		i++;
		}
	*pP2 = '\0';
	pP++;
	iVerBeta = atoi(buf2);
	pP2 = buf2;	

	return(iVerMajor * 10000 + iVerMinor * 100 + iVerBeta);
	}



