/*---------------------------------------------------------------------------+
|
|     (C)opyright Dennis Bareis and Andreas Buchinger
|     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
|
|     MODULE NAME:   REXXPMPF.C
|
|
|         $Author:   Dennis_Bareis  $
|                    Andreas Buchinger  2009-2024
|       $Revision:   1.0  $
|           $Date:   17 Aug 1996 13:39:22  $
|        $Logfile:   C:/PROJECTS/OS2/PVCS.IT/PMPRINTF/REXXPMPF.C_V  $
|
|     DESCRIPTION:   This module contains routines to transmit data to
|                    PMPRINT.EXE.  This code can be called via REXX
|                    procedures.
+---------------------------------------------------------------------------*/

/*------ Make constant string definition more efficient & protect them! -----*/
#pragma  strings(readonly)

/*------ Include required Header files --------------------------------------*/
#define  INCL_NOPMAPI
#define  INCL_LONGLONG  // 20090705 AB

#include <os2.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#define  INCL_REXXSAA
#include <rexxsaa.h>
#include "PMPRINTF.H"


/*--- The following codes are used to abort the Rexx Procedure --------------*/
#define  RCSYNTAX_INVALID_NUMBER_OF_PARAMETERS  40
#define  RCSYNTAX_INVALID_PARM_1                40

/*--- Define some macros ----------------------------------------------------*/
#define BUILDRXSTRING(t, s)              \
        do                               \
        {                                \
           strcpy((t)->strptr,(s));      \
           (t)->strlength = strlen((s)); \
        }  while (FALSE)


/*--- Define the API calls available to rexx code ---------------------------*/
RexxFunctionHandler RexxPmpf;
RexxFunctionHandler RexxPmPrintfModeSet;
RexxFunctionHandler RexxPmPrintfQueueNameThisProcess;


/*--- We wish to initialize the mode so that its more relevant for rexx code ---*/
static BOOL PmPrintfModeIsSet = FALSE;




/*===========================================================================*/
static void SetDefaultPmprintfModeIfRequired(void)
/*===========================================================================*/
	{
	if ( ! PmPrintfModeIsSet )
		{
		PmPrintfModeIsSet = TRUE;
		PmPrintfModeSet(0);
		}
	}



/*===========================================================================*/
static ULONG _Optlink RexSetRexxVariable(char  *RexxVariableName, char  * NewValue)
/*===========================================================================*/
	{
	SHVBLOCK   VarPoolCtrlBlock;
	long       UpdateVarRc;

	MAKERXSTRING(VarPoolCtrlBlock.shvname,  RexxVariableName, strlen(RexxVariableName));
	MAKERXSTRING(VarPoolCtrlBlock.shvvalue, NewValue,         strlen(NewValue));
	VarPoolCtrlBlock.shvvaluelen = strlen(NewValue);
	VarPoolCtrlBlock.shvcode     = RXSHV_SYSET;					//Set the variable
	VarPoolCtrlBlock.shvret      = (char )0;						//Clear return code field
	VarPoolCtrlBlock.shvnext     = (SHVBLOCK *)0;				//No next block

	/*--- Create/Update the variable ----------------------------------------*/
	UpdateVarRc = RexxVariablePool(&VarPoolCtrlBlock);

	/*--- We don't want specific return code for variable creation so reset -*/
	if ( UpdateVarRc == RXSHV_NEWV )
		UpdateVarRc = RXSHV_OK;

	/*--- Return the result -------------------------------------------------*/
	return( UpdateVarRc );
	}


/*===========================================================================*/
static ULONG _Optlink RexGetRexxVariable(char  *RexxVariableName, char  FetchCode, char  * Result, ULONG ResultBufferLength)
/*===========================================================================*/
	{
	SHVBLOCK   VarPoolCtrlBlock;
	ULONG      GetVarRc;

	VarPoolCtrlBlock.shvcode            = FetchCode;			 //Get the variable (from where?)
	VarPoolCtrlBlock.shvvalue.strptr    = NULL;						 //Get rexx to allocate memory
	VarPoolCtrlBlock.shvvalue.strlength = 0;
	MAKERXSTRING(VarPoolCtrlBlock.shvname,  RexxVariableName, strlen(RexxVariableName));
	VarPoolCtrlBlock.shvvaluelen        = 0;
	VarPoolCtrlBlock.shvret             = (char )0;				 //Clear return code field
	VarPoolCtrlBlock.shvnext            = (SHVBLOCK *)0;	//No next block

	/*--- Create/Update the variable -----------------------------------------*/
	if ( (GetVarRc=RexxVariablePool(&VarPoolCtrlBlock)) != 0 )
		{
		/*--- Clear buffer & return the reason -------------------------------*/
		*Result = '\0';
		return(GetVarRc);
		}

	/*--- Return the result --------------------------------------------------*/
	if ( VarPoolCtrlBlock.shvvalue.strlength == 0 )
		{
		*Result = '\0';
		return(0);
		}
	else
		{
		if ( VarPoolCtrlBlock.shvvalue.strlength < ResultBufferLength )
			{
			/*--- Will fit in our buffer -------------------------------------*/
			memcpy(Result, VarPoolCtrlBlock.shvvalue.strptr, VarPoolCtrlBlock.shvvalue.strlength);
			Result[VarPoolCtrlBlock.shvvalue.strlength] = '\0';
			return(0);
			}
		else
			{
			/*--- Will NOT fit in our buffer (truncate) ----------------------*/
			memcpy(Result, VarPoolCtrlBlock.shvvalue.strptr, ResultBufferLength-1);
			Result[ResultBufferLength-1] = '\0';
			return(RXSHV_TRUNC);
			}
		}
	}






/*===========================================================================*/
ULONG RexxPmPrintfModeSet( CHAR     * name,
													 ULONG      NumberOfParameters,
													 RXSTRING   args[],
													 CHAR     * queuename,
													 RXSTRING * ReturnCode
												 )

/*                                                                           */
/*  This procedure sets the Mode that PMPRINTF.EXE works under.              */
/*===========================================================================*/
	{
	PMPRINTF_MODE   NewMode;
	ULONG           RoutineRc;
	char            StringRoutineRc[20];
	char          * StringValue;
	char          * Ptr;

	/*--- We expect one only parameter ---------------------------------------*/
	SetDefaultPmprintfModeIfRequired();
	if ( NumberOfParameters != 1 )
		return(RCSYNTAX_INVALID_NUMBER_OF_PARAMETERS);

	/*--- Convert the passed string value to ULONG ---------------------------*/
	StringValue = args[0].strptr;
	if ( (StringValue[0] == '0') && (StringValue[1] == 'x') )
		NewMode = strtoul(StringValue+2, &Ptr, 16);
	else
		NewMode	= strtoul(StringValue, &Ptr, 10);
	if ( *Ptr != '\0' )
		return(RCSYNTAX_INVALID_PARM_1);		 //Syntax error, invalid mode

	/*--- Now tell PMPRINTF what Mode to use ---------------------------------*/
	RoutineRc = PmPrintfModeSet(NewMode);

	/*--- Return the previous mode to the caller -----------------------------*/
	sprintf(StringRoutineRc, "%lu", RoutineRc);
	BUILDRXSTRING(ReturnCode, StringRoutineRc);

	/*--- return OK to rexx (rexx script will not die!) ----------------------*/
	PmPrintfModeIsSet = TRUE;
	return(0);
	}




/*===========================================================================*/
ULONG RexxPmPrintfQueueNameThisProcess(    CHAR     * name,
																					 ULONG      NumberOfParameters,
																					 RXSTRING   args[],
																					 CHAR     * queuename,
																					 RXSTRING * ReturnCode
																			)
/*                                                                           */
/*  This procedure sets the queue name which will be used when talking to    */
/*  PMPRINTF.EXE                                                             */
/*===========================================================================*/
	{
	char  * StaticQueueName;
	ULONG   RoutineRc;
	char    StringRoutineRc[20];

	/*--- We expect one only parameter ---------------------------------------*/
	if ( NumberOfParameters != 1 )
		return(RCSYNTAX_INVALID_NUMBER_OF_PARAMETERS);

	/*--- Now tell PMPRINTF what Queue Name to use ---------------------------*/
	if ( (StaticQueueName=strdup(args[0].strptr)) == NULL )
		RoutineRc = __LINE__;
	else
		{
		/*--- We successfully duplicated string so tell PMPRINTF to use it ---*/
		if ( PmPrintfQueueNameThisProcess(StaticQueueName) != NULL )
			RoutineRc = 0;
		else
			{
			/*--- Must be invalid queue name ---------------------------------*/
			return(RCSYNTAX_INVALID_PARM_1);		 //Syntax error, invalid queue name
			}
		}

	/*--- Set up the return code so the caller knows what happened (if he cares!) ---*/
	sprintf(StringRoutineRc, "%lu", RoutineRc);
	BUILDRXSTRING(ReturnCode, StringRoutineRc);

	/*--- return OK to rexx (rexx script will not die!) ----------------------*/
	return(0);
	}



/*===========================================================================*/
ULONG RexxPmpf(  CHAR     * name,
								 ULONG      NumberOfParameters,
								 RXSTRING   args[],
								 CHAR     * queuename,
								 RXSTRING * ReturnCode
							)

/*                                                                           */
/*  This procedure expects a single parameter from the rexx caller, this is  */
/*  sent to PMPRINTF.EXE                                                     */
/*===========================================================================*/
	{
	ULONG   PfRc;
	char    StringPfRc[20];
	char    SourceInfo[CCHMAXPATH + 50];		 //Allow a bit for preceeding garbage
	char  * SourceCmd;
	char    CallLineNumber[20];

	/*--- We expect one only parameter ---------------------------------------*/
	SetDefaultPmprintfModeIfRequired();
	if ( NumberOfParameters != 1 )
		return(RCSYNTAX_INVALID_NUMBER_OF_PARAMETERS);

	/*--- Get the short name of the rexx procedure ---------------------------*/
	if ( RexGetRexxVariable("SOURCE", RXSHV_PRIV, SourceInfo, sizeof(SourceInfo)) != 0 )
		SourceCmd = "?.CMD";
	else
		{
		if ( (SourceCmd=strrchr(SourceInfo, '\\')) == NULL )
			SourceCmd = SourceInfo;
		else
			SourceCmd++;							 //Point past the '\' character
		}

	/*--- Get information about source of last call --------------------------*/
	RexGetRexxVariable("SIGL", RXSHV_SYFET, CallLineNumber, sizeof(CallLineNumber));
	if ( CallLineNumber[0] == '\0' )
		strcpy(CallLineNumber, "0");

	/*--- Now Send the passed string to PMPRINTF.EXE -------------------------*/
	PfRc = PmPrintf("%s,*%s: %s", SourceCmd, CallLineNumber, args[0].strptr);

	/*--- Set up the return code so the caller knows what happened (if he cares!) ---*/
	sprintf(StringPfRc, "%lu", PfRc);
	BUILDRXSTRING(ReturnCode, StringPfRc);

	/*--- return OK to rexx (rexx script will not die!) ----------------------*/
	return(0);
	}


