/*  REXX script to generate a buildlevel string. If file 'outfile' (see below)
    is found it will be updated with new values defined below. Otherwise it
    will be created. Also modifies PMPRINTF.DEF and WPSDEBUG.DEF

    Uses enviroment variable %hostname% as build machine entry
    and current build date/time. Version is taken from FECOMMON.h.

    ---------------------------------------------------------------------------
    Arguments:
      ARG(1) when set 1 or 2 outputs additional messages usefull for debuging
      this script
    ---------------------------------------------------------------------------
    This file is part of PmPrintf package for OS/2, ArcaOS, eComStation
    Copyright (c) 2009-2024, Andreas Buchinger
    ---------------------------------------------------------------------------
    History:
    20231227 AB   initial, modified from wpa_supplicant
*/

/* Use value from environment (no more than 31 chars) or default below */
VendorName = value('VENDOR',,'OS2ENVIRONMENT')
/*                                     |--- no more than 31 chars ---| */
if VendorName == '' THEN VendorName = 'Dennis Bareis,Andreas Buchinger'

BuildMachine = value('hostname',,'OS2ENVIRONMENT')
ASD      = ''           /* empty string by default for wpa_supplicant */
Language = ''           /* f.i. EN or empty string */
Country  = ''           /* f.i. AT or empty string */
CPU      = ''           /* U for uniprocessor or empty string */
FixPack  = ''           /* */
/*             |-----------  Description should not be longer than 79 characters  ------------|     */
Description = 'PmPrintf - printf() like debug statements to PM viewer or log file'
ver_years = '1996-2024'
ver_rev =   ''          /* leave empty as it's taken from FECOMMON.h */
ver_build = ''          /* leave empty as it's taken from FECOMMON.h (empty (default) or numeric value) */
/*ver_min ='GA2'*/             /* use f.i. beta or leave empty */
ver_min =''             /* leave empty as it's taken from FECOMMON.h (use f.i. beta or leave empty) */

outfile1  = 'bldlevel.def'
outfile2  = 'PMPRINTF.DEF'
outfile3  = 'WPSDEBUG.DEF'

/* some internally used variables */
CrLf     = '0d0a'x;
BldLevelStartTag = '@#';
BldLevelEndTag   = '#@';
Redirection  = '> NUL 2>&1';
fTrace = 0

/* process command line parameters
called with parameter (fTrace) 1 outputs debug messages
called with parameter (fTrace) 2 outputs even more debug info   */
PARSE ARG fTrace

/* limit vendor name */
IF LENGTH(VendorName) > 31 THEN DO
    VendorName = LEFT(VendorName,31)
END

/* get version info from FECOMMON.h */
file = 'FECOMMON.h'
IF fTrace > 0 THEN SAY 'Take version info from file='file
IF (lines(file) = 0) THEN SAY '- error:' file 'could not be opened.';

DO WHILE lines(file) > 0
    line = LINEIN(file)
    line = STRIP(line)
    IF LEFT(line, 2) == '//' THEN ITERATE
    IF LEFT(line, 2) == '/*' THEN ITERATE
    pos = WORDPOS('PGM_VERSION ' , line)
    IF pos > 0 THEN DO
        line=TRANSLATE(line, ' ', '"')
        ver_maj = WORD(line, pos + 1)
        IF fTrace > 1 THEN SAY 'Version(ver_maj)='ver_maj
        /* do not process rest of the file */
        LEAVE
    END
END

/* Append years to description string */
Description = Description' 'ver_years

IF WORDLENGTH(ver_rev, 1) > 0 THEN ver_maj = ver_maj'.'ver_rev

IF WORDLENGTH(ver_maj, 1) > 10 THEN ver_maj = LEFT(ver_maj, 10)
IF fTrace > 0 THEN SAY 'Version(ver_maj)='ver_maj

/* limit max. build string length */
IF LENGTH(ver_build) > 7 THEN ver_build = LEFT(ver_build,7)

/* compose date and time strings */
datetime = DATE('S')
datetime = INSERT('-', datetime ,6)
datetime = INSERT('-', datetime ,4)
time = TIME()
time = TRANSLATE(time, ' ', ':')
hour = WORD(time, 1)
minu=  WORD(time, 2)
IF LENGTH(hour) < 2 THEN hour = INSERT('0',hour)
IF LENGTH(minu) < 2 THEN minu = INSERT('0',minu)
time = hour':'minu
/* datetime = year'-'month'-'day' 'time */
datetime = datetime' 'time
IF fTrace > 1 THEN SAY 'datetime='datetime

/* compose Feature string (:ASD:EN:US:4b:U:8101) */
Feature=':'ASD':'Language':'Country':'ver_build':'CPU':'FixPack
IF fTrace > 1 THEN  SAY 'Feature='Feature

/* build time stamp und build machine string */
/* date/time have to be 26 chars (leading ' ' required) */
IF LENGTH(BuildMachine) > 11 THEN DO
    BuildMachine = LEFT(BuildMachine, 11)
    END
bldmachlen = LENGTH(BuildMachine)
IF fTrace > 1 THEN SAY 'bldmach='BuildMachine'<-- len='bldmachlen
datimmach = LEFT(datetime, 24)
datimmach = datimmach' 'BuildMachine
IF fTrace > 1 THEN SAY datimmach'<--'

/* bldlevel does not display more than 79 characters */
/* 2024xx AB limit to 39 cause of unknown problem starting some exe otherwise */
/* see emails with Steven */
IF LENGTH(Description) > 39 THEN DO
    Description = LEFT(Description, 39)
    SAY 'Description lenght corrected'
    END

IF (ver_min \= '') THEN DO
  FullString = "@#"VendorName':'ver_maj'.'ver_min'#@##1## 'datimmach''Feature'@@  'Description""
  END
ELSE DO
  FullString = "@#"VendorName':'ver_maj'#@##1## 'datimmach''Feature'@@  'Description""
  END

IF fTrace > 0 THEN SAY FullString

/*
 * Replace DESCRIPTION string in .def file
 *
 */

Do I=1 to 3 by 1
  File = VALUE(outfile''I);

  IF (STREAM( File, 'C', 'OPEN READ') \= 'READY:') THEN
  DO
     SAY '- error:' File 'could not be opened.';
     rc = ERROR.ACCESS_DENIED;
  END;

  NewFileContents = '';
  DescriptionOkay = FALSE;

  DO WHILE (LINES( File) > 0)
     ThisLine = LINEIN( File);
     IF fTrace > 1 THEN IF I = 1 THEN SAY ThisLine

     IF (TRANSLATE( WORD( ThisLine, 1)) = 'DESCRIPTION') THEN
     DO
        IF fTrace > 0 THEN SAY '^--- description found in file='File
        ThisLine = "DESCRIPTION '"FullString"'"
     END;

     /* contents for new file */
     NewFileContents = NewFileContents''CrLf''ThisLine;
     DescriptionOkay = TRUE;
  END;
  IF (DescriptionOkay \= TRUE) THEN NewFileContents = NewFileContents''CrLf''"DESCRIPTION '"FullString"'"


  DO
     /* close and rewrite file */
     rcx = STREAM( File, 'C', 'CLOSE');

     /* write tmp file */
     Contents = SUBSTR( NewFileContents, 3);
     TmpFile = SysTempFilename( VALUE( 'TMP',,OS2ENVIRONMENT)'\bld_temp.???');
     IF (LINEOUT( TmpFile, Contents)) THEN
     DO
        SAY '- error: cannot write' TmpFile;
        EXIT( 99);
     END;
     rcx = LINEOUT( TmpFile);

     /* copy over file */
     '@COPY' TmpFile File Redirection;
     IF (rc \= 0) THEN
     DO
        SAY '- error: cannot write' File;
        EXIT( 99);
     END;
     rcx = SysFileDelete( TmpFile);

  END;

END;

Do I=1 to 3 by 1
  File = outfile''I;
  IF (File \= '') THEN rcx = STREAM( File, 'C', 'CLOSE');
  END

RETURN( ERROR.NO_ERROR);
